import numpy as np, pylab as pl
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import PolynomialFeatures

def bootstrapped_model(xs, ys, degree, n_bootstraps=3000):
    #xs = np.linspace(0,1,100)
    #ys = xs**2 + np.random.normal(0,.05,size=100)


    pl.scatter(xs, ys, color='blue', marker='.', s=1, label='Data')

    #xs = np.array(xs); ys = np.array(ys)
    #xs = xs[np.isfinite(ys)]; ys = ys[np.isfinite(ys)]
    #xs = xs[np.argsort(xs)]; ys = ys[np.argsort(xs)]

    x = np.atleast_2d(xs).T
    y = np.array(ys)



    xdomain = np.linspace(xs.min(), xs.max(), 1000)

    predictions = np.zeros((n_bootstraps, len(xdomain)))

    for i in range(n_bootstraps):

        indices = np.random.choice(len(x), len(x), replace=True)

        model = LinearRegression()
        poly = PolynomialFeatures(degree=degree, include_bias=True)
        poly_features = poly.fit_transform(x[indices])

        model.fit(poly_features, y[indices])

        predictions[i, :] = model.predict(poly.fit_transform(np.atleast_2d(xdomain).T))  # Save the predictions for each bootstrap sample

    alpha = 0.05  # Set the significance level
    lower_ci = np.percentile(predictions, 100 * alpha / 2, axis=0)
    upper_ci = np.percentile(predictions, 100 * (1 - alpha / 2), axis=0)

    mean_pred = np.mean(predictions, axis=0)

    pl.ylim(y.min(), y.max())

    pl.plot(xdomain, mean_pred, color='k', label='mean')
    pl.fill_between(xdomain, lower_ci, upper_ci, color='gray', alpha=0.2,
                    label='95% predictive interval')
    pl.legend(loc='best')


def test_basics():

    n = 1000
    xs = np.random.uniform(-5, 5, size=n)
    ys = xs * np.cos(xs*5) + np.random.normal(0,.5,size=n)

    bootstrapped_model(xs, ys, degree=20)
    pl.show()




if __name__ == '__main__':
    from arsenal import testing_framework
    testing_framework(globals())




class assert_throws(object):
    """
    Contextmanager + Decorator for asserting that a certain exceptions (or no
    exception) will arise with it's context.

    As a context manager:

      >>> with assert_throws(ZeroDivisionError):
      ...     1/0

      >>> with assert_throws(None):
      ...     pass

      >>> with assert_throws(None, ZeroDivisionError):
      ...     pass

      >>> with assert_throws(ZeroDivisionError):
      ...     pass
      Traceback (most recent call last):
          ...
      AssertionError: did not raise required ZeroDivisionError. Got None instead.

      >>> with assert_throws(AssertionError, ZeroDivisionError):
      ...     pass
      Traceback (most recent call last):
          ...
      AssertionError: did not raise required AssertionError or ZeroDivisionError. Got None instead.

    """

    def __init__(self, *expect):
        self.expect = expect
        super(assert_throws, self).__init__()

    def __enter__(self, *_):
        pass

    def __exit__(self, exc, val, _):
        msg = ' or '.join('None' if e is None else e.__name__ for e in self.expect)

        if exc is None:
            if None not in self.expect:
                raise AssertionError('did not raise required %s. Got %s instead.' % (msg, None))
        else:
            if not issubclass(exc, self.expect):
                raise AssertionError('did not raise required %s. Got %s instead.' % (msg, exc.__name__))

        return True


if __name__ == '__main__':
    import doctest
    doctest.testmod()

    def test():

        with assert_throws(Exception):
            print(1/0)

        with assert_throws(ZeroDivisionError):
            print(1/0)

        with assert_throws(None):
            pass

        try:
            with assert_throws(ZeroDivisionError, TypeError, ValueError):
                pass
        except AssertionError:
            pass
        else:
            raise AssertionError('test failed.')

        print('Passed basic tests.')

    test()



from arsenal.iterextras import buf_iter, head_iter
from heapq import heapify, heappop, _siftup, heappush


def sorted_union(*iterators):
    """
    Merge multiple sorted inputs into a single sorted output.

    Equivalent to `sorted(itertools.chain(*iterables))`, but more efficient and lazy.

    >>> list(merge_sorted([1,3,5,7], [0,2,4,8], [5,10,15,20], [], [25]))
    [0, 1, 2, 3, 4, 5, 5, 7, 8, 10, 15, 20, 25]

    """

    h = [head_iter(s) for s in iterators]
    h = [s for s in h if not s.done]
    heapify(h)
    while h:
        s = h[0]
        yield s.__next__()       # advance the top iterator
        if s.done:
            heappop(h)           # remove empty iterator
        else:
            _siftup(h, 0)        # restore heap condition

merge_sorted = sorted_union


class Item:
    def __init__(self, cost, index, elems):
        self.cost = cost
        self.index = index
        self.elems = elems
    def __lt__(self, other):
        return self.cost < other.cost


# Note: If the product operation is associative, it is best to pick an
# association order rather than use the k-fold product.  It will speed up the
# computation by a factor of k: the k-fold products in the inner loop below
# don't cost O(k) because the effective k is equal to 2.
def sorted_product(p, *iters):
    """
    Sorted product of `iters`, where the output is sorted by a monotonic product
    operator `p`.  Examples: tuples, or multiplication/addition of positive numbers.
    """

    n = len(iters)
    assert n > 1

    # use buffered iterator to ensure (random) access to the previously emitted
    # values of each iterator.
    iters = [buf_iter(it) for it in iters]

    def vals(z):
        return tuple(it[j] for it, j in zip(iters, z))

    # elements in the heap are wrapped `Item` to make it a min heap.
    q = []
    y = (0,)*n

    try:
        heappush(q, Item(p(vals(y)), 0, y))
    except IndexError:
        return   # empty

    while q:
        item = heappop(q)
        x = item.elems
        j = item.index

        yield item.cost

        # next best item must differ by one, enqueue all such items
        # We reduce the number of pushes by the dotted rule trick.
        for i in range(j, n):

            y = list(x)
            y[i] = x[i] + 1
            y = tuple(y)

            #if x[i] + 1 >= len(a[i]): continue
            try:
                iters[i][y[i]]
            except IndexError:
                # `IndexError` is thrown when `iter[i]` is finite and we
                # requested more iterates than it has.
                continue

            heappush(q, Item(p(vals(y)), i, y))


def main():
    import numpy as np
    import itertools
    from arsenal.iterextras import take
    from itertools import count


    # weighted tuples are the idea as a path weight with backpointers; our weighted
    # tuple copies the tuple, so it is inefficient compared to the lazier
    # backpointer variant.
    class WeightedTuple:
        def __init__(self, w, *key):
            self.key = key
            self.w = w
        def __lt__(self, other):
            return (self.w, self.key) < (other.w, other.key)
        def __eq__(self, other):
            return (self.w, self.key) == (other.w, other.key)
        def __mul__(self, other):
            return LWeightedTuple(self.w*other.w, self, other)
        def __add__(self, other):
            return LWeightedTuple(self.w+other.w, self, other)
        def __iter__(self):
            return iter((self.w, self.key))
        def __repr__(self):
            return repr((self.w, self.key))


    class LWeightedTuple(WeightedTuple):
        "WeightedTuple with lazy concatenation of keys."
        def __init__(self, w, a, b):
            self.w = w
            self.a = a
            self.b = b
        @property
        def key(self):
            return self.a.key + self.b.key


    def wprod(xs):
        return np.product([WeightedTuple(x, x) for x in xs])

    def wsum(xs):
        return np.sum([WeightedTuple(x, x) for x in xs])

    def check(iters):
        for p in [np.product, np.sum, tuple, wprod]:
            # enumerate and sort; not lazy
            want = list(sorted(p(x) for x in itertools.product(*iters)))
            have = list(sorted_product(p, *iters))
            print()
            print('product operator:', p.__name__)
            print('HAVE:', have)
            #if have != want:
            print('WANT:', want)
            assert have == want
        print('pass.')

    print('===========')
    check([
        (.1, .4, 0.5),
        (0.09, 0.11, 0.8),
        (0.111, .3, 0.6),
    ])
    print('===========')
    check([
        (1, 2, 3),
        (4, 7, 11),
    ])
    print('===========')
    check([
        (0.01, .4, 0.5),
        (0.11, 0.8),
        (0.6,),
    ])
    print('===========')
    check([
        (1, 2, 3, 100),
        (4, 7, 9),
        (14, 17, 19),
        (24, 27, 29),
    ])
    print('===========')


    a = (3**i for i in count(1))
    b = (4**i for i in count(1))
    c = (5**i for i in count(1))

    for s,x in take(20, sorted_product(wsum, a, b, c)):
        print(s, x)


if __name__ == '__main__':
    main()



"""
`PartitionRefinement` is a data structure for representing a partition of a
set.  Unlike the better-known `UnionFind` data structure, which *merges* sets,
the `PartitionRefinement` data structure performs the "dual" operation of
*splitting* sets.

Important use cases:
  - Efficient DFA minimization

References:
  https://en.wikipedia.org/wiki/Partition_refinement

Original version by Ryan Cotterell

"""

import numpy as np
from collections import defaultdict
from arsenal import colors


def stable(f, P):
    "Is partition P stable?"
    D = {}
    for n, B in enumerate(P):
        for q in B:
            D[q] = n
    for B in P:
        for p in B:
            for q in B:
                if D[f[p]] != D[f[q]]:
                    return False
    return True


def split(S, P):
    return frozenset(P & S), frozenset(P-S)


def hopcroft(f, P):

    # compute the pre-image of f
    finv = defaultdict(set)
    for n in f:
        finv[f[n]].add(n)

    stack = list(P)

    while stack: # empties in O(log n) steps
        S = stack.pop()
        R = set() # new refinement

        # compute subset of the pre-image in O(n) time
        Sinv = set.union(*[finv[x] for x in S])

        for B in P: # entire loop runs in O(n) time
            X, Y = split(Sinv, B) # runs in O(|B|) time

            if len(X) > 0 and len(Y) > 0:
                # X, Y are now part of the refinement
                R.add(X)
                R.add(Y)

                # Hopcroft's speed-up to the slower algorithm is that we
                # only need to enqueue the smaller set.
                if len(X) < len(Y):
                    stack.append(X)
                else:
                    stack.append(Y)
            else:
                # Q remains part of the refinement
                R.add(B)
        P = R

    return frozenset(P)


def slow(f, P):

    # compute the pre-image of f
    finv = defaultdict(set)
    for n in f:
        finv[f[n]].add(n)

    stack = list(P)

    while stack: # empties in O(n) steps
        S = stack.pop()
        R = set() # new refinement

        # compute subset of the pre-image in O(n) time
        Sinv = set.union(*[finv[x] for x in S])

        for B in P: # entire loop runs in O(n) time
            X, Y = split(Sinv, B) # runs in O(|B|) time

            if len(X) > 0 and len(Y) > 0:
                # X, Y are now part of the refinement
                R.add(X)
                R.add(Y)

                # X, Y become future splitters
                stack.append(X)
                stack.append(Y)
            else:
                # Q remains part of the refinement
                R.add(B)
        P = R

    return frozenset(P)


def test_partition():
    import random

    for _ in range(25):
        N = 50

        # create random total function
        f = dict(zip(range(N), random.choices(range(N), k=N)))

        # create random partition
        i = random.randint(2, N-1)
        P = np.random.permutation(range(N))
        P = { frozenset(P[:i]), frozenset(P[i:]) }

        want = slow(f, P)
        have = hopcroft(f, P)
        assert want == have
        assert stable(f, have)
        #print(want)

    print(colors.ok)


if __name__ == '__main__':
    from arsenal import testing_framework
    testing_framework(globals())



class Alphabet(object):
    """
    Class for maintaining a perfect hash for a set of keys.

    >>> a = Alphabet()
    >>> [a[x] for x in 'abcd']
    [0, 1, 2, 3]

    >>> list(map(a.lookup, range(4)))
    ['a', 'b', 'c', 'd']

    >>> a.freeze()
    >>> a.add('z')
    Traceback (most recent call last):
      ...
    ValueError: Alphabet is frozen. Key "z" not found.

    >>> print(a.plaintext())
    a
    b
    c
    d

    >>> print(a)
    Alphabet(size=4,frozen=True)

    >>> list(a)
    ['a', 'b', 'c', 'd']

    >>> a == Alphabet(['a', 'b', 'c', 'd'])
    True

    >>> a == Alphabet(['b', 'a', 'c', 'd'])
    False

    >>> a.map('aabc')
    [0, 0, 1, 2]

    """

    def __init__(self, data=()):
        self._map = {}       # str -> int
        self._list = []      # int -> str
        self._frozen = False
        self.add_many(data)

    def __repr__(self):
        return 'Alphabet(size=%s,frozen=%s)' % (len(self), self._frozen)

    def freeze(self):
        self._frozen = True

#    def keys(self):
#        return self._map.keys()

    def items(self):
        return self._map.items()

    def imap(self, seq):
        """
        Apply alphabet to sequence while filtering. By default, `None` is not
        emitted, so the Note that the output sequence may have fewer items.
        """
        for s in seq:
            yield self[s]

    def map(self, seq):
        return list(self.imap(seq))

    def add_many(self, x):
        for k in x:
            self.add(k)

    def lookup(self, i):
        return self._list[i]

    def lookup_many(self, x):
        return list(map(self.lookup, x))

    def __contains__(self, k):
        return k in self._map

    def __getitem__(self, k):
        try:
            return self._map[k]
        except KeyError:
            #if not isinstance(k, basestring):
            #    raise ValueError("Invalid key (%s): only strings allowed." % (k,))
            if self._frozen:
                raise ValueError('Alphabet is frozen. Key "%s" not found.' % (k,))
            x = self._map[k] = len(self._list)
            self._list.append(k)
            return x

    __call__ = __getitem__
    add = __getitem__

    def __iter__(self):
        return iter(self._list)

    def __len__(self):
        return len(self._map)

    def plaintext(self):
        "assumes keys are strings"
        return '\n'.join(self)

    @classmethod
    def load(cls, filename):
        #if not os.path.exists(filename): return cls()
        with open(filename) as f:
            return cls(l.strip() for l in f)

    def save(self, filename):
        with open(filename, 'w') as f:
            f.write(self.plaintext())

    def __eq__(self, other):
        return self._list == other._list

    encode_many = map
    decode_many = lookup_many

    encode = __getitem__
    decode = lookup


if __name__ == '__main__':
    import doctest
    doctest.testmod()



import matplotlib.pyplot as pl
from numpy import sqrt, mean, array
from numpy.linalg import svd
from mpl_toolkits.mplot3d import Axes3D


def mds(X, dimensions=2):
    """
    Multidimensional Scaling - Given a matrix of interpoint distances,
    find a set of low dimensional points that have similar interpoint
    distances.

    Author: Jeremy M. Stober
    with modifications by Tim Vieira
    """
    E = (-0.5 * X * X)   # square the distances
    # Use mat to get column and row means to act as column and row means.
    Er = array(mean(E, axis=1))
    Es = array(mean(E, axis=0))
    # From Principles of Multivariate Analysis: A User's Perspective (pg 107).
    F = array(E - Er.T - Es + mean(E))
    # svd to get low rank approximation or F
    [U, S, _] = svd(F)
    Y = U * sqrt(S)
    return (Y[:,0:dimensions], S)


def mds_scatter(distance):
    Y, _ = mds(distance)
    pl.scatter(Y[:,0], Y[:,1], s=20, c='b', marker='o')
    pl.grid(True)
    return Y


def mds_plot2D(points, distance):
    Y, _ = mds(distance)
    pl.figure(figsize=(12, 6))
    pl.scatter(Y[:,0], Y[:,1], s=20, c='b', marker='o')
    for user, x, y in zip(points, Y[:,0], Y[:,1]):
        pl.text(x, y, user, fontsize=12)
    pl.grid(True)
    return Y


def mds_plot3D(points, distance):
    Y, _ = mds(distance, 3)
    ax = Axes3D(pl.figure())
    ax.scatter(Y[:,0], Y[:,1], Y[:,2])
    list(map(ax.text, Y[:,0], Y[:,1], Y[:,2], points))  # add labels to points



"""
Enumerate alignments between two sets
"""

def alignments(x, y):
    "Enumerate alignments between two sets"
    yield from _alignments(x, y, 0, set(range(len(y))), [])


def _alignments(
        x, y,
        cursor: 'cursor into x',
        remaining: 'available indices in y',
        align: 'current alignment prefix'
):
    if len(align) == len(x):
        yield align
        return

    # XXX: Below are some heuristics to short-circuit backtracking sooner.
#    from collections import Counter
#    xx = Counter(x[cursor:])
#    yy = Counter([y[j] for j in remaining])
#    for a in xx:
#        if xx[a] > yy[a]:  # if we don't have enough `a`s in y to cover x, quit now.
#            return

    if cursor >= len(x): return
    if len(remaining) == 0: return
    for j in remaining:
        if x[cursor] == y[j]:
            yield from _alignments(x, y, cursor+1,
                                   remaining - {j},
                                   align + [(cursor,j)])


def check_property(x, y, a):
    for i,j in a:
        assert x[i] == y[j]
        ii, jj = zip(*a)
        # test that `a` is a one-to-one function
        assert len(set(ii)) == len(x) == len(set(jj)) == len(a)
        assert set(ii) == set(range(len(x)))
        assert set(jj) <= set(range(len(y)))


def test():
    from arsenal import ok
    x, y = 'abc', 'xxabcxabcabcaaax'

    aligns = []
    for t, a in enumerate(alignments(x, y)):
        check_property(x, y, a)
        aligns.append(frozenset(a))
    assert len(set(aligns)) == len(aligns)   # check for spurious ambiguity
    assert len(aligns) == 6*3*3

    x, y = 'cba', 'xabcc'
    aligns = []
    for t, a in enumerate(alignments(x, y)):
        check_property(x, y, a)
        aligns.append(frozenset(a))
    assert len(set(aligns)) == len(aligns)   # check for spurious ambiguity
    assert len(aligns) == 2

    print(ok)


if __name__ == '__main__':
    test()



import numpy as np
import matplotlib.pyplot as pl
import pandas as pd
from scipy.stats import pearsonr, spearmanr
from scipy.linalg import lstsq, norm
from arsenal import colors
from arsenal.iterview import progress
from arsenal.maths import cosine, linf, relative_difference, zero_retrieval
from arsenal.maths.rvs import cdf
from arsenal import Alphabet


def pp_plot(a, b, pts=100, show_line=True):
    "P-P plot for samples `a` and `b`."
    A = cdf(a)
    B = cdf(b)
    x = np.linspace(min(a.min(), b.min()), max(a.max(), b.max()), pts)

    if show_line:
        pl.plot(range(0,2), range(0,2), ':', alpha=0.5, c='k', lw=2)

    pl.plot(A(x), B(x), alpha=1.0)


def check_equal(want, have, verbose=0, **kwargs):
    c = compare(want, have, verbose=verbose, **kwargs)
    if c.max_relative_error > 0.01:
        assert False, c.format_message()
    return c


def align(X, Y, distance=lambda x,y: abs(x - y), maximize=False):
    """
    Find a cheap alignment of X and Y.
    """
    from scipy.optimize import linear_sum_assignment
    n = len(X); m = len(Y)
    c = np.zeros((n, m))
    for i, x in enumerate(X):
        for j, y in enumerate(Y):
            c[i,j] = distance(x, y)

    rs,cs = linear_sum_assignment(c, maximize=maximize)

    class result:
        cost = c[rs,cs].sum()
        x = list(X[r] for r in rs)
        y = list(Y[c] for c in cs)
        x_ind = rs
        y_ind = cs
        cost_matrix = c

    return result


class compare:

    def __init__(self, want, have, name=None, data=None, P_LARGER=0.9,
                 regression=True, ax=None, alphabet=None,
                 want_label=None, have_label=None, verbose=1):
        """Compare vectors.

        Arguments:

          - Specifying data for comparison two methods:

            1) `want`, `have`: two numeric one-dimensional arrays which we'd like
               to compare (the argument names come for software testing). This
               method requires argument `data=None`.

            2) `data`: instance of `DataFrame`, expects arguments `want` and `have`
               to be column labels.

          - `name`: name of this comparison.

        Note:

         - when plotting `want` is the y-axis, `have` is the x-axis. This is by
           convention that `want` is the dependent variable (regression target).

        TODO:

         - Add an option to drop NaNs and continue comparison.

         - Indicate which dimensions have the largest errors.

         - Add option to allow alignment/matchings?

        """

        self.name = name

        if isinstance(alphabet, Alphabet):
            alphabet = list(alphabet)

        if isinstance(want, dict) and isinstance(have, dict):
            alphabet = list(want.keys() | have.keys()) if alphabet is None else alphabet
            #assert set(have.keys()) == set(alphabet), \
            #    'Keys differ.\n  have keys  = %s\n  want keys = %s' % (set(have.keys()), set(alphabet))
            want = [want.get(k, 0) for k in alphabet]
            have = [have.get(k, 0) for k in alphabet]

        if isinstance(want, np.ndarray) and isinstance(have, np.ndarray):
            assert alphabet is None
            alphabet = Alphabet(dict(np.ndenumerate(want)).keys())

            assert want.shape == have.shape, [want.shape, have.shape]
            want = want.flatten()
            have = have.flatten()

        if data is not None:
#            assert isinstance(want, (int, str)), \
#                'expected a column name have %s' % type(want)
#            assert isinstance(have, (int, str)), \
#                'expected a column name have %s' % type(have)

            if want_label is None: want_label = want
            if have_label is None: have_label = have

            want = data[want]
            have = data[have]

        else:
            if want_label is None: want_label = 'want'
            if have_label is None: have_label = 'have'

            want = np.asarray(want)
            have = np.asarray(have)

            data = pd.DataFrame({want_label: want, have_label: have})

        assert want.shape == have.shape, [want.shape, have.shape]
        [n] = want.shape

        self.want = want
        self.have = have
        self.alphabet = alphabet
        self.ax = ax
        self.have_label = have_label
        self.want_label = want_label
        self.n = n
        self.coeff = None

        self.tests = tests = []

        # Check that vectors are finite.
        if not np.isfinite(want).all():
            tests.append(['want finite', progress(np.isfinite(want).sum(), n), False])
        if not np.isfinite(have).all():
            tests.append(['have finite', progress(np.isfinite(have).sum(), n), False])

        ne = norm(want)
        ng = norm(have)
        ok = abs(ne-ng)/ne < 0.01 if ne != 0 else True

        if n > 1:
            tests.append(['norms', '[%g, %g]' % (ne, ng), ok])
            #F = zero_retrieval(want, have)
            #tests.append(['zero F1', F, F > 0.99])

        # Correlation statistics
        if n > 1:
            #self.cosine = cosine(want, have)
            #tests.append(['cosine', self.cosine, (self.cosine > 0.99999)])   # cosine similarities must be really high.
            self.pearson = 1.0 if ne == ng == 0 else pearsonr(want, have)[0]
            tests.append(['pearson', self.pearson, (self.pearson > 0.99999)])
            self.spearman = spearmanr(want, have)[0]
            tests.append(['spearman', self.spearman, (self.spearman > 0.99999)])

        # TODO: this check should probably take into account the scale of the data.
        d = linf(want, have)
        self.max_err = d
        tests.append(['ℓ∞', d, None])
        tests.append(['ℓ₂', np.linalg.norm(want - have), None])

        # same sign check (weak agreement, but useful sanity check -- especially
        # for gradients)
        x = want
        y = have
        s = np.asarray(~((x >= 0) ^ (y >= 0)), dtype=int)
        p = s.sum() * 100.0 / len(s)
        tests.append(['same-sign', f'{p:.2f}% ({s.sum()}/{len(s)})', p == 100.0])

        # relative error
        r = relative_difference(want, have)
        r = np.max(r[np.isfinite(r)])
        #tests.append(['max rel err', r, r <= 0.01])
        self.max_relative_error = r
        self.max_rel_err = r

        # TODO: suggest that if relative error is high and rescaled error is low (or
        # something to do wtih regression residuals) that maybe there is a
        # (hopefully) simple fix via scale/offset.

        # TODO: can provide descriptive statistics for each vector
        #tests.append(['range (want)', [want.min(), want.max()], 2])
        #tests.append(['range (have)   ', [have.min(), have.max()], 2])

        # regression and rescaled error only valid for n >= 2
#        if n >= 2:
#            es = abs(want).max()
#            gs = abs(have).max()
#            if es == 0:
#                es = 1
#            if gs == 0:
#                gs = 1
#            if 0:
#                # rescaled error
#                E = want / es
#                G = have / gs
#                R = abs(E - G)
#                r = np.mean(R)
#                tests.append(['mean rescaled error', r, r <= 1e-5])

        if regression:
            self.regression()

        if n >= 2:
            # These tests check if one of the datasets is consistently larger than the
            # other. The threshold for error is based on `P_LARGER` ("percent larger").
            L = ((want-have) > 0).sum()
            if L >= P_LARGER * n:
                tests.append(['want is larger', progress(L, n), 0])
            L = ((have-want) > 0).sum()
            if L >= P_LARGER * n:
                tests.append(['have is larger', progress(L, n), 0])

        self.tests = tests
        if verbose:
            self.message()

    def message(self):
        print()
        print(self.format_message())
        print()

    def format_message(self):
        lines = [
            'Comparison%s: n=%s' % (' (%s)' % self.name if self.name else '', self.n),
        ]
        for k, v, passed in self.tests:
            if passed == 1:
                c = colors.green
            elif passed == 0:
                c = colors.red
            else:
                c = colors.yellow
            try:
                v = '%g' % v
            except TypeError:
                pass
            lines.append('  %s: %s' % (k, c % (v,)))
        return '\n'.join(lines)

    def plot(self, regression=True, seaborn=False, ax=None, title=None, name=None, **scatter_kw):
        if ax is not None:
            self.ax = ax

        title = name or self.name or title

#        if seaborn:
#            import seaborn as sns
#            sns.set_context(rc={"figure.figsize": (7, 5)})
#            g = sns.JointGrid(self.have_label, self.want_label, data=self.data)
#            g.plot(sns.regplot, sns.distplot, spearmanr)
#            print("Pearson's r: {0}".format(self.pearson))
#        else:
        if 1:
            if self.ax is None:
                self.ax = pl.figure().add_subplot(111)

            self.ax.scatter(self.have, self.want, lw=0, alpha=0.5, **scatter_kw)

            self.ax.set_xlabel(self.have_label)
            self.ax.set_ylabel(self.want_label)

            # Keeps the plot region tight against the data (allow 5% of the
            # data-range for padding so that points in the scatter plot aren't
            # partially clipped.)
#            xeps = 0.05 * np.ptp(self.have)
#            self.ax.set_xlim(self.have.min() - xeps, self.have.max() + xeps)

#            yeps = 0.05 * np.ptp(self.want)
#            self.ax.set_ylim(self.want.min() - yeps, self.want.max() + yeps)

        if title is not None: self.ax.set_title(title)
        if regression: self.regression_line()
        return self

    def show(self, *args, **kw):
        self.plot(*args, **kw)
        pl.show()
        return self

    def live_plot(self, *args, **kw):
        from arsenal.viz import axman
        name = self.name if self.name is not None else kw.get('name')
        assert name is not None, '`name` required for `live_plot` in order to make the figure.'
        with axman(name) as ax:
            kw['ax'] = ax
            self.plot(*args, **kw)
        return self

    def regression_line(self):
        # TODO: write the coeff to plot.
        if self.coeff is not None and np.isfinite(self.coeff).all():
            xa, xb = self.ax.get_xlim()
            A = np.ones((self.n, 2))
            A[:,0] = self.have
            # plot estimated line
            ys = A @ self.coeff
            self.ax.plot(A[:,0], ys, c='r', alpha=0.5)
            self.ax.grid(True)
            self.ax.set_xlim(xa,xb)
        return self

    def regression(self):
        "least squares linear regression"
        if self.n < 2:
            return
        # data can't contain any NaNs
        if not np.isfinite(self.have).all() or not np.isfinite(self.want).all():
            self.tests.append(['regression',
                               'did not run due to NaNs in data',
                               0])
            return

        A = np.ones((self.n, 2))
        A[:,0] = self.have

        [self.coeff, residues, _, _] = lstsq(A, self.want)
        self.R = float(np.sqrt(residues)) if residues else 0.0

        # Label with warn or ok.
        ok = 1 if abs(self.coeff - [1, 0]).max() <= 1e-5 else 2

        slope = float(self.coeff[0])
        intercept = float(self.coeff[1])
        self.tests.append([
            'regression',
            f'[{slope:.3f} {intercept:.3f}] R={self.R:.3f}',
            ok
        ])

    def show_errors(self):
        "show largest relative errors"

        df = []

        def dot_aligned(x):
            sx = [('%g' % s) for s in x]
            dots = [s.find('.') for s in sx]
            m = max(dots)
            y = [' '*(m - d) + s for s, d in zip(sx, dots)]
            z = max(map(len, y))
            fmt = '%%-%ss' % z
            z = [fmt % s for s in y]
            return z

        s_want = dot_aligned(self.want)
        s_have = dot_aligned(self.have)

        if self.alphabet is None:
            self.alphabet = range(self.n)

        for (i,(x,y,sx,sy)) in enumerate(zip(self.want, self.have, s_want, s_have)):
            # XXX: skip zeros.
            #if abs(x) < 1e-10 and abs(y) < 1e-10:
            #    e = 0.0

            e = relative_difference(x, y)

            if e <= 0.001:
                continue

            df.append([e, self.alphabet.lookup(i), x, y, sx, sy])

            #df.append({'name':   alphabet[i],
            #           'error':  e,
            #           'want': x,
            #           'have':    y})

        df.sort(reverse=1)

        if len(df):
            print(' Largest errors')
            print(' ===============')
            for e, n, x, y, sx, sy in df:

                types = []
                if x < y:
                    types.append('bigger')
                else:
                    types.append('smaller')

                # highlight sign errors.
                #if np.sign(x) != np.sign(y):
                if (x >= 0) != (y >= 0):
                    types.append(colors.red % 'wrong sign')

                print('  %-15s %.2f%%   %s  %s' % (n, e, sx, sy), \
                    ((colors.green % 'ok') if e <= 0.01 else colors.red % 'bad'), \
                    '(%s)' % (', '.join(types)))

        #from pandas import DataFrame
        #df = DataFrame(df)
        #df.set_index('name', inplace=1)
        #print df.sort('error', ascending=0)
        return self


if __name__ == '__main__':

    def test_compare():
        n = 100
        # `a` is a noisy version of `b`, but tends to overestimate.
        a = np.linspace(0,1,n)
        b = a + np.random.uniform(-0.01, 0.1, size=n)
        compare(a,b).show()
        compare('a', 'b', data=pd.DataFrame({'a': a, 'b': b})).show()

        n = 10000
        a = np.random.uniform(-0.1, 0.1, size=n)
        b = np.random.uniform(-0.2, 0.1, size=n)
        pl.figure()
        pp_plot(a, b)
        pl.show()

    test_compare()



"""
Evaluations methods common in NLP and information extraction.


TODO: Have a look at
  https://github.com/nschneid/pyutil/blob/master/chunkeval.py, there appear to be
  richer evaluation methods.

"""


from collections import defaultdict


class F1:

    def __init__(self, confusion_matrix=False):
        if confusion_matrix:
            self.confusion_matrix = defaultdict(lambda : defaultdict(int))
        else:
            self.confusion_matrix = None
        self.n_examples = 0
        self.relevant  = defaultdict(set)
        self.retrieved = defaultdict(set)

    def report(self, instance, prediction, target):
        if self.confusion_matrix is not None:
            self.confusion_matrix[target][prediction] += 1
        self.n_examples += 1
        self.add_relevant(target, instance)
        self.add_retrieved(prediction, instance)

    def add_relevant(self, label, instance):
        self.relevant[label].add(instance)

    def add_retrieved(self, label, instance):
        self.retrieved[label].add(instance)
        return instance in self.relevant[label]

    def latex(self):
        relevant  = self.relevant
        retrieved = self.retrieved

        print(r"""
\begin{tabular}{|l|c|c|c|c|}
\hline
 Label         &   Count   &   Precision   &   Recall   &   $F_1$   \\
\hline""")

        tbl = []
        labels = list(self.relevant.keys())
        labels.sort()
        for label in labels:
            R = P = F = 0
            count = len(relevant[label])
            top = relevant[label] & retrieved[label]
            if len(relevant[label]) != 0:
                R = len(top) / len(relevant[label])
            if len(retrieved[label]) != 0:
                P = len(top) / len(retrieved[label])
            if P + R != 0:
                F = 2*P*R / (P + R)
            print(r' %8s & %5d & %5.1f & %5.1f & %5.1f \\' % (label, count, P*100, R*100, F*100))
            tbl.append((label,count,P,R,F))

        print(r"""\hline
\end{tabular}
""")
        return tbl

    def scores(self, verbose=True):
        relevant  = self.relevant
        retrieved = self.retrieved
        if verbose:
            m = max(list(map(len, self.relevant))) if self.relevant else 0
            fmt = ' | %{0}s | %5d | %5.1f | %5.1f | %5.1f |'.format(m)

            line = ' |' + '='*m + '==================================|'

            print(line)
            print(' |', ' '*m, '|   C   |   P   |   R   |   F   |')
            print(line)

        tbl = []
        labels = list(self.relevant.keys())
        labels.sort()
        for label in labels:
            R = P = F = 0
            count = len(relevant[label])
            top = relevant[label] & retrieved[label]
            if len(relevant[label]) != 0:
                R = len(top) / len(relevant[label])
            if len(retrieved[label]) != 0:
                P = len(top) / len(retrieved[label])
            if P + R != 0:
                F = 2*P*R / (P + R)
            if verbose:
                print(fmt % (label, count, P*100, R*100, F*100))
                #t.add_row([label, P*100, R*100, F*100])
            tbl.append((label,count,P,R,F))
        if verbose:
            print(line)
            #print t
        return tbl

    def confusion(self):
        assert self.confusion_matrix is not None
        for t, predictions in self.confusion_matrix.items():
            incorrect = sum(cnt for p,cnt in list(predictions.items()) if t != p)
            print('%s [correct: %s; incorrect: %s]' % (t, predictions[t], incorrect))
            for predicted, cnt in predictions.items():
                if t != predicted:
                    incorrect += cnt
                    print('      %6s -> %s' % (predicted, cnt))
            print()


def plot_confusion(y_true, y_pred, alphabet, normalized=False):
    """
    Draw confusion matrix

    Options:

     - normalized: Normalize the confusion matrix by row (i.e by the number of samples in each
       class)

    """
    import numpy as np
    import matplotlib.pyplot as pl

    from sklearn.metrics import confusion_matrix

    def plot_confusion_matrix(cm, title='Confusion matrix', cmap=pl.cm.Blues):
        pl.imshow(cm, interpolation='nearest', cmap=cmap)
        pl.title(title)
        pl.colorbar()
        target_names = list(alphabet)
        tick_marks = np.arange(len(target_names))
        pl.xticks(tick_marks, target_names, rotation=45)
        pl.yticks(tick_marks, target_names)
        pl.tight_layout()
        pl.ylabel('True label')
        pl.xlabel('Predicted label')

    # Compute confusion matrix
    cm = confusion_matrix(y_true, y_pred)
    np.set_printoptions(precision=2)

    if normalized:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

    print(cm)

    pl.figure()
    plot_confusion_matrix(cm, title='Confusion matrix')

    pl.show()



import heapq
from operator import getitem, itemgetter
from itertools import chain, count, cycle, islice, repeat, tee
from random import shuffle
from collections import defaultdict
from arsenal.iterview import iterview


SENTINEL = object()
class head_iter:
    def __init__(self, iterator):
        self.done = False
        self.head = SENTINEL
        self.tail = iter(iterator)
        self.__next__()
    def __lt__(self, other):
        if self.done:
            return True
        if other.done:
            return False
        if self.head == SENTINEL:
            return True
        if other.head == SENTINEL:
            return False
        return self.head < other.head
    def __next__(self):
        h = self.head
        try:
            self.head = self.tail.__next__()
        except StopIteration:
            self.done = True
            self.head = SENTINEL
        return h


class buf_iter:
    "Lazy, random access wrapper around an iterator."
    def __init__(self, x):
        self.buf = []
        self.x = iter(x)
        self.j = 0     # num things emitted

    def __iter__(self):
        return self

    def __next__(self):
        try:
            y = self[self.j]
            self.j += 1
            return y
        except IndexError:
            raise StopIteration()

    def __len__(self):
        for _ in self: pass
        return len(self.buf)

    def __getitem__(self, i):
        k = i - len(self.buf)
        while k >= 0:
            try:
                self.buf.append(self.x.__next__())
            except StopIteration:
                raise IndexError(i)
            k -= 1
        return self.buf[i]


def argmax(f, seq):
    """
    >>> argmax(lambda x: -x**2 + 1, range(-10,10))
    0
    """
    return argmax2(f,seq)[1]


def argmax2(f, seq):
    """
    >>> argmax2(lambda x: -x**2 + 1, range(-10,10))
    (1, 0)
    """
    return max(((f(x),x) for x in seq), key=itemgetter(0))


def argmin(f, seq):
    """
    >>> argmin(lambda x: x**2 + 1, range(-10,10))
    0
    """
    return argmin2(f,seq)[1]


def argmin2(f, seq):
    """
    >>> argmin2(lambda x: x**2 + 1, range(-10,10))
    (1, 0)
    """
    return min(((f(x),x) for x in seq), key=itemgetter(0))


def groupby2(s, key=lambda x: x):
    """
    Eager version of groupby which does what you'd expect groupby to do.

    >>> dict(groupby2(range(10), lambda x: x % 2))
    {0: [0, 2, 4, 6, 8], 1: [1, 3, 5, 7, 9]}
    """
    groups = defaultdict(list)
    for x in s:
        groups[key(x)].append(x)
    return groups


def atmost(k, seq):
    """
    >>> atmost(1, [0,0,0])
    True

    >>> atmost(1, [0,1,0])
    True

    >>> atmost(1, [0,1,1])
    False
    """
    cnt = 0
    if cnt > k:
        return False
    for x in seq:
        if x:
            cnt += 1
            if cnt > k:
                return False
    return True


def flatten(xs, types=(list, tuple)):
    if isinstance(xs, types):
        for x in xs:
            yield from flatten(x, types)
    else:
        yield xs


def partition(data, proportion):
    """
    Deterministically partition `data` according to proportion.
    Note: assumes sum(proportion) <= 1.0

    >>> partition(range(10), [0.3, 0.7])
    [[0, 1, 2], [3, 4, 5, 6, 7, 8, 9]]
    """
    assert(sum(proportion) <= 1)
    data = list(data)
    D = iter(data)
    N = len(data)
    return [list(take(j, D)) for j in [int(p*N) for p in proportion]]


def breadth_first(tree, children=iter, depth=-1, queue=None):
    """Traverse the nodes of a tree in breadth-first order.

    - (No need to check for cycles.)

    - The first argument should be the tree root;

    - children should be a function taking as argument a tree node and returning
      an iterator of the node's children.

    """
    if queue is None:
        queue = []
    queue.append(tree)
    while queue:
        node = queue.pop(0)
        yield node
        if depth != 0:
            try:
                queue += children(node)
                depth -= 1
            except (StopIteration, TypeError, IndexError):
                pass


def compress(data, selectors):
    """
    >>> list(compress('ABCDEF', [1,0,1,0,1,1]))
    ['A', 'C', 'E', 'F']
    """
    return (d for d, s in zip(data, selectors) if s)


def k_fold_cross_validation(X, K, randomize=False):
    """
    Generates K (training, validation) pairs from the items in X.

    Each pair is a partition of X, where validation is an iterable
    of length len(X)/K. So each training iterable is of length (K-1)*len(X)/K.

    If randomise is true, a copy of X is shuffled before partitioning,
    otherwise its order is preserved in training and validation.


    >>> for train, test in k_fold_cross_validation(range(3), 3):
    ...     print('test:', test, ' train:', list(train), sep=' ')
    test: [0]  train: [1, 2]
    test: [1]  train: [0, 2]
    test: [2]  train: [0, 1]

    """
    if randomize:
        X = list(X)
        shuffle(X)

    # folds = [X[i::K] for i in xrange(K)]

    # run thru X once and splitting into folds
    folds = [[] for k in range(K)]
    for i, item in enumerate(X):
        folds[i % K].append(item)

    for k in range(K):
        training = chain(*(fold for j, fold in enumerate(folds) if j != k))
        yield training, folds[k]


def unique(iterable, key=None):
    """ List unique elements, preserving order. Remember all elements ever seen. """
    # unique_everseen('AAAABBBCCDAABBB') --> A B C D
    # unique_everseen('ABBCcAD', str.lower) --> A B C D
    seen = set()
    seen_add = seen.add
    if key is None:
        for element in iterable:
            if element not in seen:
                seen_add(element)
                yield element
    else:
        for element in iterable:
            k = key(element)
            if k not in seen:
                seen_add(k)
                yield element


def window(iterable, k):
    """
    s -> (s[0],...,s[k]) (s[1],...,s[k+1]) ... (s[i],...,s[i+k]) ... (s[n-1-k],...,s[n-1])

    >>> [x+y+z for x,y,z in window('abcdef', 3)]
    ['abc', 'bcd', 'cde', 'def']

    """
    iterators = tee(iterable, k)
    for i, it in enumerate(iterators):
        # advance iterator, 'it',  by i steps
        for _ in range(i):
            try:
                next(it)    # advance by one
            except StopIteration:
                pass
    return zip(*iterators)


#_______________________________________________________________________________
#

def drop(iterator, n):
    "Advance the iterator n-steps ahead."
    # advance to the emtpy slice starting at position n
    next(islice(iterator, n, n), None)


def take(n, seq):
    """ Return the first n items in a sequence. """
    return islice(seq, None, n)


def padnone(seq):
    """ Returns the sequence elements and then returns None indefinitely """
    return chain(seq, repeat(None))


def ncycles(seq, n):
    """
    Returns the sequence elements n times.
    """
    return chain(*repeat(seq, n))


def batch(size, iterable):
    """Yield a list of (up to) batchsize items at a time.  The last
    element will be shorter if there are items left over.
    batch(s, 2) -> [s0,s1], [s2,s3], [s4, s5], ...

    >>> list(batch(2, range(5)))
    [[0, 1], [2, 3], [4]]
    """
    current = []
    for item in iterable:
        current.append(item)
        if len(current) == size:
            yield current
            current = []
    if current:
        yield current


if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
Utilities for working with Jupyter notebooks.
"""

from inspect import getsource
from IPython.display import HTML, display


def psource(*functions):
    """
    Print the source code for the given function(s).

    Based on https://github.com/aimacode/aima-python/blob/master/notebook.py

    """
    source_code = '\n\n'.join(getsource(fn) for fn in functions)
    try:
        from pygments.formatters import HtmlFormatter
        from pygments.lexers import PythonLexer
        from pygments import highlight

        display(HTML(highlight(source_code, PythonLexer(), HtmlFormatter(full=True))))

    except ImportError:
        print(source_code)


def format_source(*functions):
    """
    Print the source code for the given function(s).

    Based on https://github.com/aimacode/aima-python/blob/master/notebook.py

    """
    source_code = '\n\n'.join(getsource(fn) for fn in functions)
    try:
        from pygments.formatters import HtmlFormatter
        from pygments.lexers import PythonLexer
        from pygments import highlight

        return highlight(source_code, PythonLexer(), HtmlFormatter(full=True))

    except ImportError:
        print(source_code)



import sys
from time import time

__all__ = ['iterview']


def iterview(
        x: 'Iterable',
        msg: 'Prefix message' = None,
        #mintime: 'show progress at most every `mintime` seconds' = 0.25,
        length: 'length hint; required when len(x) fails.' = None,
        width: 'max width of progress bar' = 78,
        show: 'do not show progress bar if False' = True,
        transient = False,
):

    if not show: return x

    from rich.progress import track
    return track(x,
                 description=msg,
                 disable=not show,
                 transient=transient,
                 total=length)


def progress(n, length):
    """
    Returns a string indicating current progress.
    """
    p = 100 * n / length if length > 0 else 0
    return ('%5.1f%% (%*d/%d)'
            % ((p, len(str(length)), n, length)))


#def progress_bar(max_width, n, length):
#    """
#    Returns a progress bar (string).
#
#    Arguments:
#
#    max_width -- maximum width of the progress bar
#    """
#
#    width = int((float(n) / length) * max_width + 0.5) # at least one
#
#    if max_width - width:
#        spacing = '>' + (' ' * (max_width - width))[1:]
#    else:
#        spacing = ''
#
#    return '[%s%s]' % ('=' * width, spacing)
#
#
#def time_remaining(elapsed, n, length):
#    """
#    Returns a string indicating the time remaining (if not complete)
#    or the total time elapsed (if complete).
#    """
#
#    if n == 0:
#        return '--:--:--'
#
#    if n == length:
#        seconds = int(elapsed) # if complete, total time elapsed
#    else:
#        seconds = int((elapsed / n) * (length - n)) # otherwise, time remaining
#
#    minutes, seconds = divmod(seconds, 60)
#    hours, minutes = divmod(minutes, 60)
#
#    return '%02d:%02d:%02d' % (hours, minutes, seconds)
#
#
#def time_elapsed(elapsed):
#    """
#    Returns a string indicating the time elapsed.
#    """
#
#    seconds = int(elapsed) # if complete, total time elapsed
#    minutes, seconds = divmod(seconds, 60)
#    hours, minutes = divmod(minutes, 60)
#
#    return '%02d:%02d:%02d' % (hours, minutes, seconds)
#

#def fmt(start, n, length, width, done=False):
#    """
#    returns ...
#    """
#
#    string = progress(n, length) + ' ' # current progress
#
#    if n == length or done:
#        end = ' '
#    else:
#        end = ' ETA '
#
#    if done:
#        end += time_elapsed(time() - start)
#    else:
#        end += time_remaining(time() - start, n, length)
#
#    string += progress_bar(width - len(string) - len(end), n, length)
#    string += end
#
#    return string
#
#
#def _iterview(
#        x: 'Iterable',
#        msg: 'Prefix message' = None,
#        mintime: 'show progress at most every `mintime` seconds' = 0.25,
#        length: 'length hint; required when len(x) fails.' = None,
#        width: 'max width of progress bar' = 78,
#        show: 'do not show progress bar if False' = True
#):
#    """
#    Show a progress bar as we move through the iterable `x`.
#
#    >>> for i in iterview(range(10000), msg='foo'): pass
#    foo  20.0% ( 2001/10000) [=========>                                  ] 00:00:02
#
#    """
#
#    if not show:
#        yield from x
#
#    else:
#        start = time()
#
#        if length is None:
#            try:
#                length = len(x)
#            except TypeError:
#                raise TypeError(f'Iterable `{x!r}` does not have a known length.\n'
#                                'Please provide in an `x` where `len(x)` is defined '
#                                'or explicitly provide a value in the `length` argument.')
#        else:
#            length = int(length)
#
#        if length == 0:
#            return
#
#        n = 0
#        if msg:
#            assert len(msg) + 1 <= width, 'msg too wide'
#            width = width - len(msg) - 1
#            msg = msg + ' '
#        else:
#            msg = ''
#
#        last_update = 0
#
#        try:
#            for n, y in enumerate(x):
#                if not mintime or time() - last_update >= mintime:
#                    sys.stderr.write('\r%s%s' % (msg, fmt(start, n, length, width)))
#                    last_update = time()
#                yield y
#        finally:
#            sys.stderr.write('\r%s%s\n' % (msg, fmt(start, n+1, length, width, done=1)))


def tests():
    #from arsenal.assertions import assert_throws
    from time import sleep

    # Check for error
    #with assert_throws(ValueError):
    #    list(iterview((None for _ in range(5))))

    # won't throw an error because length is passed in
    list(iterview((None for _ in range(5)), length=5))

    for _ in iterview(range(10000), msg='foo',
                      #mintime=0.25
                      ):
        sleep(0.0001)

    # Print time elapsed if we terminate earlier than expected.
    for i in iterview(range(10000), msg='foo',
                      #mintime=0.25
                      ):
        if i == 2000: break
        sleep(0.001)

    from arsenal import terminal
    print('should disappear', terminal.arrow.down)
    # Print time elapsed if we terminate earlier than expected.
    for i in iterview(range(10000), msg='foo', #mintime=0.25,
                      transient=True):
        if i == 2000: break
        sleep(0.001)
    print('should disappear', terminal.arrow.up)



if __name__ == '__main__':
    tests()



import numpy as np
from scipy.linalg import inv, det
from scipy.linalg.blas import dger


class Inv:

    def __init__(self, A):
        B = np.array(inv(A), order='F', copy=True)
        assert B.flags['F_CONTIGUOUS']
        self.B = B

    def rank_one_update(self, u, v):
        # https://timvieira.github.io/blog/post/2021/03/25/fast-rank-one-updates-to-matrix-inverse/
        B = self.B
        Bu = B @ u
        s = 1 + float(v.T @ Bu)
        alpha = -1 / s
        # Warning: `overwrite_a=True` silently fails when B is not an order=F array!
        dger(alpha, Bu, v.T @ B, a=B, overwrite_a=1)
        return s

    @property
    def value(self):
        return self.B


class InvAndDet(Inv):

    def __init__(self, A):
        super().__init__(A)
        self.D = det(A)

    def rank_one_update(self, u, v):
        s = super().rank_one_update(u, v)
        self.D *= s


def test():

    n = 10

    A = np.random.randn(n,n)
    B = inv(A)
    u = np.random.randn(n,1)
    v = np.random.randn(n,1)

    B = InvAndDet(A)

    B.rank_one_update(u,v)
    assert np.allclose(inv(A + np.outer(u, v)), B.value)

    print(det(A + np.outer(u, v)), B.D)
    assert np.allclose(det(A + np.outer(u, v)), B.D)

    print('ok')


if __name__ == '__main__':
    test()



import re, os, sys, traceback, warnings, webbrowser
import subprocess, tempfile, http.server
from functools import wraps
from io import StringIO
from contextlib import contextmanager
from arsenal import colors

# TODO: stdio/stderr redirection is available in contextlib; so is exception
# suppression and even the context decorator.


def open_diff(a, b, cmd='meld'):
    "View diff of string representations in dedicated diff program."
    print(a, file=open('/tmp/a', 'w', encoding='utf-8'))
    print(b, file=open('/tmp/b', 'w', encoding='utf-8'))
    os.system('%s /tmp/a /tmp/b' % cmd)


#def colordiff(x, y):
#    print(colors.nocolor(x), file=open('/tmp/x','w', encoding='utf-8'))
#    print(colors.nocolor(y), file=open('/tmp/y','w', encoding='utf-8'))
#    cmd = 'colordiff -t --color=always --side-by-side --ignore-space-change /tmp/x /tmp/y'
#    result = subprocess.run(cmd.split(), stdout=subprocess.PIPE)
#    return result.stdout.decode('utf-8')


def deprecated(use_instead=None, msg=None):
    """
    This is a decorator which can be used to mark functions as deprecated.
    It will result in a warning being emitted when the function is used.
    """
    def wrapped(func):
        @wraps(func)
        def new_func(*args, **kwargs):
            m = f'Call to deprecated function `{func.__name__}`.'
            if msg:
                m += f'\n{msg}'
            if use_instead:
                m += f"\nUse `{use_instead}` instead."
            warnings.warn(m, stacklevel=2)
            return func(*args, **kwargs)
        return new_func

    return wrapped


def trace(f):
    def wrapper(*args):
        trace.print(f'{f.__name__}({", ".join(map(repr, args))})')
        trace.indent += 1
        result = f(*args)
        trace.indent -= 1
        trace.print(f' -> {result}')
        return result
    return wrapper
trace.indent = 0
trace.spaces = '    '
def _trace(*msg):
    tab = trace.spaces*trace.indent
    if trace.on: print(f'{tab}{" ".join(map(str, msg))}')
trace.print = _trace
trace.on = True


class ddict(dict):
    """
    Variation on collections.defaultdict which allows default value callback to
    inspect missing key.
    """
    def __init__(self, f):
        self.f = f
        super(ddict, self).__init__()

    def __missing__(self, key):
        self[key] = c = self.f(key)
        return c


@contextmanager
def ignore_error(color='red'):
    try:
        yield
    except:
        etype, evalue, tb = sys.exc_info()
        tb = '\n'.join(traceback.format_exception(etype, evalue, tb))

        if color is not None:
            color = getattr(colors, color)
        else:
            color = '%s'

        print(color % '*'*80)
        print(color % tb)
        print(color % '*'*80)


#@deprecated
#class logger(object):
#
#    def __init__(self, filename, clean=True, quiet=False):
#        if clean:
#            os.system('rm -f ' + filename)
#        self.f = file(filename, 'wb')
#        self.quiet = quiet
#
#    def write(self, x):
#        self.f.write(x)
#        if not self.quiet:
#            sys.__stdout__.write(x)
#
#    @staticmethod
#    def start(*args, **kw):
#        """ redirect stdout and stderr to logger """
#        log = logger(*args, **kw)
#        sys.stderr = sys.stdout = log
#        return log


def force(g):
    """ force evaluation of generator `g`. """
    @wraps(g)
    def wrap(*args, **kw):
        return list(g(*args, **kw))
    return wrap


#def attn(msg):
#    """ Display a dialog which is sure to get my attention. """
#    import gtk
#    window = gtk.Window(gtk.WINDOW_TOPLEVEL)
#    window.connect("delete_event", lambda *x: False)
#    window.connect("destroy", lambda *x: gtk.main_quit())
#    window.set_border_width(10)
#    button = gtk.Button(msg)
#    button.connect_object("clicked", gtk.Widget.destroy, window)
#    window.add(button)
#    button.show()
#    window.show()
#    window.fullscreen()
#    gtk.main()


## # borrowed from whoosh
## def find_object(name, blacklist=None, whitelist=None):
##     """Imports and returns an object given a fully qualified name.
##
##     >>> find_object("whoosh.analysis.StopFilter")
##     <class 'whoosh.analysis.StopFilter'>
##     """
##
##     if blacklist:
##         for pre in blacklist:
##             if name.startswith(pre):
##                 raise TypeError("%r: can't instantiate names starting with %r" % (name, pre))
##     if whitelist:
##         passes = False
##         for pre in whitelist:
##             if name.startswith(pre):
##                 passes = True
##                 break
##         if not passes:
##             raise TypeError("Can't instantiate %r" % name)
##
##     lastdot = name.rfind(".")
##
##     assert lastdot > -1, "Name %r must be fully qualified" % name
##     modname = name[:lastdot]
##     clsname = name[lastdot + 1:]
##
##     mod = __import__(modname, fromlist=[clsname])
##     cls = getattr(mod, clsname)
##     return cls


def piped():
    """ Returns piped input via stdin, else None. """
    return sys.stdin if not sys.stdin.isatty() else None


def highlighter(p, flags=0):
    pattern = re.compile('%s' % p, flags)
    return lambda x: pattern.sub(colors.bold % colors.yellow % colors.bg_red % r'\1', x)


def browser(html):
    """
    Display html in the default web browser without creating a temp file.

    Instantiates a trivial http server and calls webbrowser.open with a URL
    to retrieve html from that server.
    """

    class RequestHandler(http.server.BaseHTTPRequestHandler):
        def do_GET(self):
            bufferSize = 1024*1024
            for i in range(0, len(html), bufferSize):
                self.wfile.write(html[i:i+bufferSize])

    server = http.server.HTTPServer(('127.0.0.1', 0), RequestHandler)
    webbrowser.open('http://127.0.0.1:%s' % server.server_port)
    server.handle_request()


def pager(s, cmd='less'):
    """Use the pager passed in and send string s through it."""
    subprocess.Popen(cmd, stdin=subprocess.PIPE).communicate(s)


def edit_with_editor(s=None):
    """
    Open os.environ['EDITOR'] and load in text s.

    Returns the text typed in the editor, after running strip().
    """
    # This is the first time I've used with!
    with tempfile.NamedTemporaryFile() as t:
        if s:
            t.write(str(s))
            t.seek(0)
        subprocess.call([os.environ.get('EDITOR', 'nano'), t.name])
        return t.read().strip()

editor = edit_with_editor

#_________________________________________________________________________
#

# TODO: use a contextdecorator
@contextmanager
def ctx_redirect_io(f=None):
    r"""
    Usage example:
      >>> with ctx_redirect_io() as io_target:
      ...    print('how is this for io?')
      >>> io_target.getvalue()
      'how is this for io?\n'
    """
    target = f or StringIO()

    # Redirect IO to target.
    original_stdout = sys.stdout
    try:
        sys.stdout = target
        # provide an entry point for the procedure we are wrapping as well
        # as a reference to target
        yield target
    finally:
        # Restore stdio
        sys.stdout = original_stdout

redirect_io = ctx_redirect_io

#def redirect_io(f):
#    """
#    redirect all of the output to standard out to a StringIO instance,
#    which can be accessed as an attribute of the function, f.io_target
#
#    Usage Example:
#        >>> @redirect_io
#        ... def foo(x):
#        ...    print(x)
#        >>> foo('hello?')
#        >>> print(foo.io_target.getvalue())    # doctest:+NORMALIZE_WHITESPACE
#        hello?
#        >>>
#    """
#    @wraps(f)
#    def wrap(*args, **kwargs):
#        with ctx_redirect_io() as io_target:
#            wrap.io_target = io_target
#            return f(*args, **kwargs)
#    return wrap


if __name__ == '__main__':

    import doctest

    def run_tests():

        def test_redirect_io():
            msg = 'hello there?'
            with redirect_io() as f:
                print(msg)
            assert str(f.getvalue().strip()) == msg

        test_redirect_io()

    run_tests()
    print('passed')

    doctest.testmod()

    #attn('ATTENTION!')



import matplotlib.pyplot as pl
import pandas as pd
import numpy as np
from sys import stderr
from time import time
from contextlib import contextmanager
from arsenal.humanreadable import htime
from arsenal import colors
from arsenal.misc import ddict
from scipy.stats import mannwhitneyu


class Benchmark(dict):

    def __init__(self, title=None):
        self.title = title
        super().__init__()

    def __missing__(self, name):
        self[name] = c = Timer(name)
        return c

    def compare(self, statistic=np.median):
        if len(self) == 0:
            return
        if len(self) == 1:
            [x] = self.values()
            print(x)
            return
        best = min(self.values(), key=lambda t: statistic(t.times))
        for name in sorted(self):
            if name != best.name:
                best.compare(self[name], statistic=statistic)

    def plot_feature(self, feature, timecol='timer', ax=None, **kw):
        if ax is None: ax = pl.figure().add_subplot(111)
        for t in list(self.values()):
            t.plot_feature(feature=feature,
                           timecol=timecol,
                           ax=ax,
                           **kw)
        if self.title is not None: ax.set_title(self.title)
        ax.legend(loc=2)
        return ax

    def plot_survival(self,*args,**kwargs):
        "Show the probability each algorithm is still running."
        for _, t in sorted(self.items()):
            t.plot_survival(*args,**kwargs)

    def run(self, methods, reps):
        from arsenal import iterview, restore_random_state

        if isinstance(methods, (tuple, list)):
            methods = {m.__name__: m for m in methods}
        jobs = [
            (name, seed)
            for seed in range(reps)   # TODO: use a better strategy for picking random seeds.
            for name in methods
        ]
        np.random.shuffle(jobs)       # shuffle jobs to avoid weird ordering correlations
        for name, seed in iterview(jobs):
            with restore_random_state(seed):
                with self[name]:
                    methods[name]()


timers = Benchmark

class Timer:
    """
    >>> from time import sleep
    >>> a = Timer('A')
    >>> b = Timer('B')
    >>> with a:
    ...     sleep(0.5)
    >>> with b:
    ...     sleep(1)
    >>> a.compare(b)          #doctest:+SKIP
    A is 2.0018x faster

    """
    def __init__(self, name=None):
        self.name = name
        self.times = []
        self.features = []
        self.b4 = None

    def __enter__(self):
        self.b4 = time()

    def __exit__(self, *_):
        self.times.append(time() - self.b4)

    def __str__(self):
        return 'Timer(name=%s, avg=%g, std=%g)' % (self.name, self.mean, self.std)

    def __call__(self, **features):
        self.features.append(features)
        return self

    @property
    def mean(self):
        return np.mean(self.times)

    @property
    def median(self):
        return np.median(self.times)

    @property
    def std(self):
        if len(self.times) <= 1:
            return 0.0
        return np.std(self.times, ddof=1)

    @property
    def total(self):
        return sum(self.times)

    def compare(self, other, statistic=np.median, verbose=True):
        if len(self.times) == 0 or len(other.times) == 0:
            print('%s %s %s' % (self.name, '???', other.name))
            return

        self_attr = statistic(self.times)
        other_attr = statistic(other.times)
        if self_attr <= other_attr:

            # use_continuity=True, alternative=None
            # alternative = {None, ‘two-sided’, ‘less’, ‘greater’}

            # XXX: support Wilcoxon signed rank test for paired examples.
            #print(np.array(self.times) - np.array(other.times))
            U = mannwhitneyu(self.times, other.times, alternative='two-sided')

            extra = ''
            if verbose:
                pval = f'p={U.pvalue:.5f}'
                if U.pvalue < 0.05:
                    pval = colors.green % pval
                else:
                    pval = colors.yellow % pval
                extra = f'({pval}, {statistic.__name__}: {other.name}: {other_attr:g}, {self.name}: {self_attr:g})'

            print(f'{self.name} is %6.4fx faster than {other.name} %s' \
                % (other_attr / self_attr, extra))

        else:
            other.compare(self, statistic=statistic, verbose=verbose)

#    def compare_many(self, *others, **kw):
#        for x in sorted(others, key=lambda x: x.name):
#            if x != self:
#                self.compare(x, **kw)

    def plot_feature(self, feature, timecol='timer',
                     ax=None, label=None, scatter=False,
                     show_curve=False, **kw):
        if ax is None: ax = pl.figure().add_subplot(111)
        df = self.dataframe(timecol)
        a = df.groupby(feature).median()
#        a = df.groupby(feature).mean()

        X = a.index
        Y = a[timecol]
        ax.set_xlabel(feature)
        ax.set_ylabel('time (seconds)')

        if label is None: label = self.name

        [line] = ax.plot(X, Y, lw=2, alpha=0.5, label=label, **kw)
        kw.pop('c', None)

        c = line.get_color()
        ax.scatter(X, Y, c=c, lw=0, label=None, alpha=0.25, **kw)
        #ax.scatter(df[feature], df[timecol], c=c, alpha=0.25, marker='.', label=None, **kw)

        if show_curve:
            xs = np.array(X); ys = np.array(Y)
            assert np.all(xs > 0) and np.all(ys > 0)
            a,b = np.polyfit(np.log(xs), np.log(ys), deg=1)
            l = r'%s $\approx {%.2f} \cdot $%s$^{%.2f}$' % (label, np.exp(b), feature, a)
            ax.plot(xs, np.exp(b)*xs**a, alpha=0.5, label=l, c=c, linestyle=':')

        data = []
        for f, dd in df.groupby(feature):
            data.append([
                f,
                np.percentile(dd[timecol], 20),
                np.percentile(dd[timecol], 80),
            ])

            if scatter:
                ax.scatter([f]*len(dd), dd[timecol], c=c, alpha=0.25)

        data = list(sorted(data))
        fs, ls, us = zip(*data)
        ax.fill_between(fs, ls, us, alpha=0.2, color=c)

        #elif 'box' in show:
        #    # TODO: doen't work very well yet. need to fill out the x-axis since
        #    # feature might not be dense. Should throw an error if feature isn't
        #    # integral.
        #    ddd = [np.asarray(dd[timecol]) for f, dd in sorted(df.groupby(feature))]
        #    ax.boxplot(ddd)

        return ax

    def dataframe(self, timecol='timer'):
        df = pd.DataFrame(list(self.features))
        df[timecol] = self.times
        return df

    def filter(self, f, name=None):
        t = Timer(name)
        t.times, t.features = list(zip(*[(x,y) for (x,y) in zip(self.times, self.features) if f(x,y)]))
        return t

    def bucket_filter(self, feature_to_bucket, bucket_filter):
        df = self.dataframe()
        data = []
        for k, d in df.groupby(feature_to_bucket):
            d = d[bucket_filter(k, d)]
            _, x = list(zip(*d.iterrows()))
            data.append(x)
        return pd.DataFrame(data)

    def trim_slow(self, feature_to_bucket, threshold):
        return self.bucket_filter(
            feature_to_bucket,
            lambda k, d: d.timer <= d.timer.quantile(threshold)
        )

    def plot_survival(self, ax=None):
        if ax is None: ax = pl.figure().add_subplot(111)
        from arsenal.maths import cdf
        ts = np.array(self.times)
        xs = np.linspace(0, ts.max(), 1000)
        ax.plot(xs, 1-cdf(ts)(xs), label=self.name)
        ax.legend(loc='best'); ax.set_xscale('log'); ax.set_yscale('log')


@contextmanager
def timeit(name, fmt='{name} ({htime})', header=None):
    """Context Manager which prints the time it took to run code block."""
    if header is not None:
        print(header)
    b4 = time()
    yield
    sec = time() - b4
    if sec < 60:
        ht = '%.4f sec' % sec
    else:
        ht = htime(sec)
    print(fmt.format(name=name, htime=ht, sec=sec), file=stderr)


def main():
    from arsenal.iterview import iterview
    from time import sleep
    from numpy.random import uniform

    T = Benchmark('A vs B')
    for _ in iterview(range(1000), T.title):
        with T['A']:
            sleep(np.random.exponential(.001))
        with T['B']:
            sleep(np.random.exponential(.001))

    T.compare()


    t = Timer('test')

    for i in iterview(range(1, 20)):
        for _ in range(10):
            with t(i=i):
                c = 0.01
                z = max(i**2 * 0.0001 + uniform(-c, c), 0.0)
                sleep(z)

    t.plot_feature('i')
    pl.show()


if __name__ == '__main__':
    main()



import random
import numpy as np
from numpy import array, exp, log, dot, abs, multiply, cumsum, arange, \
    asarray, ones, mean, searchsorted, sqrt, isfinite, log1p, expm1
import scipy.linalg as la
from scipy import stats
from arsenal import colors
from scipy.stats import pearsonr, spearmanr
from contextlib import contextmanager
from scipy.special import expit as sigmoid


def wide_dataframe():
    import pandas as pd
    from arsenal import console_width
    pd.set_option('display.width', console_width())


@contextmanager
def set_printoptions(*args, **kw):
    "Context manager for numpy's print options."
    was = np.get_printoptions()
    np.set_printoptions(*args, **kw)
    yield
    np.set_printoptions(**was)


@contextmanager
def restore_random_state(seed=None):
    py_rng = random.getstate()
    np_rng = np.random.get_state()
    if seed is not None:
        random.seed(seed)
        np.random.seed(seed)
    yield (py_rng, np_rng)
    random.setstate(py_rng)
    assert np.random.set_state(np_rng) is None


def fit_curve(xs, ys):
    import matplotlib.pyplot as pl
    xs = np.array(xs); ys = np.array(ys)
    assert np.all(xs > 0) and np.all(ys > 0)
    a,b = np.polyfit(np.log(xs), np.log(ys), deg=1)
    label = '${%.2f} \cdot x^{%.2f}$' % (np.exp(b), a)
    print('[fit] estimate', label)
    pl.plot(xs, ys, c='r', alpha=0.5, label='data')
    pl.plot(xs, np.exp(b)*xs**a, c='b', alpha=0.5,
            label=label)
    pl.legend(loc='best')
    pl.show()


def argmin_random_tie(x):
    "argmin with randomized tie breaking."
    return np.random.choice(np.flatnonzero(x == x.min()))


def argmax_random_tie(x):
    "argmax with randomized tie breaking."
    return np.random.choice(np.flatnonzero(x == x.max()))


def onehot(i, n):
    """
    Create a one-hot vector: a vector of length `n` with a `1` at position `i`
    and zeros elsewhere.
    """
    x = np.zeros(n)
    x[i] = 1
    return x


def wls(A, b, w):
    "weighted least-squares estimation."
    from statsmodels.api import WLS
    # Note: statsmodel is a bit more accurate than directly calling lstsq
    return WLS(b, A, weights=w).fit().params


def blocks(M, k):
    """Partition the matrix M into blocks

      M = [[A, B],
           [C, D]]

    Output: `A: k x k`, `D: (n-k) x (n-k)` if `M: n x n`.

    TODO: support arbitrary partitions when `k` is a list

    """

#    if isinstance(k, list) or (isinstance(k, np.ndarray) and k.dtype == int):
#        N,M = M.shape
#        M = M[k,k]    # Shuffle the rows so that k comes first
#        nodes = set(k)
#        I = list(sorted(nodes))
#        O = list(sorted(set(range(N)) - nodes))
#        A,B,C,D = [{} for _ in range(4)]
#        for i in range(N):
#            for j in range(M):
#                v = M[i,j]
#                if i in nodes:
#                    if j in nodes:
#                        A[i,j] = v
#                    else:
#                        B[i,j] = v
#                else:
#                    if j in nodes:
#                        C[i,j] = v
#                    else:
#                        D[i,j] = v
#        k = len(k)

    assert isinstance(k, int)
    return [
        [M[:k,:k], M[:k,k:]],
        [M[k:,:k], M[k:,k:]],
    ]


class subspace:
    """
    Linear subspace
    """
    def __init__(self, A):
        # A is a collection of column vectors
        [self.dim, _] = A.shape
        self.A = A
        self.P = la.pinv(A) @ A
#        self.P = A @ la.inv(A.T @ A) @ A.T
        self.N = la.null_space(A.T)

    def project(self, q):
        return self.P @ q

    def basic_checks(self):
        # basic checks: P is a projection matrix
        P = self.P
        assert P.shape == (self.dim, self.dim)
        assert np.allclose(P, P.T)
        assert np.allclose(P @ P, P)

    def visualize(self, ax, num=100):
        if self.dim == 2:
            X = np.linspace(*ax.get_xlim(), num=num)
            a,b = self.N   # Show the linear subspace spanned by the basis vectors in F.
            ax.plot(X, -a*X/b, color='k', alpha=0.25)
        if self.dim == 3:
            X = np.linspace(*ax.get_xlim(), num=num)
            Y = np.linspace(*ax.get_ylim(), num=num)
            X,Y = np.meshgrid(X, Y)
            a,b,c = self.N   # Show the linear subspace spanned by the basis vectors in F.
            ax.plot_surface(X, Y, np.array([-(a*x+b*y)/c for x,y in zip(X.flat,Y.flat)]).reshape(X.shape),
                            color='k', alpha=0.25)
        return ax


def split_ix(N, p, randomize=1):
    "Sample a random partition of N integers with proportions `p`."
    if randomize:
        I = np.random.permutation(N)
    else:
        I = list(range(N))
    folds = []
    for q in p:
        j = int(np.ceil(N*q))
        fold = I[:j]
        folds.append(fold)
        I = I[j:]
    return folds


def norm(x, p=2):
    if not isfinite(x).all(): return np.nan
    return la.norm(x, p)


def zero_retrieval(want, have):
    "How good are we at retrieving zero values? Measured with F1 score."
    assert want.shape == have.shape
    return f1(want == 0, have == 0)


def f1(A, B):
    """
    Compute the F1 measure on two bit vectors.

    >>> f1([1], [1])
    1.0

    >>> f1([1], [0])
    0.0

    >>> f1([1,0,1], [0,1,1])
    0.5

    """

    A = np.array(A, dtype=bool); B = np.array(B, dtype=bool)

    C = (A & B).sum()
    A = A.sum()
    B = B.sum()
    R = C / A if A != 0 else 1.0
    P = C / B if B != 0 else 1.0
    F = 2*P*R/(P+R) if (P+R) != 0 else 1.0
    return F


def relative_difference(a, b):
    """Element-wise relative difference of two arrays

    $$
    \begin{cases}
    0                        & \mathrm{if } x = y = 0 \\
    |x-y| / \max(|x|, |y|)   & \mathrmPotherwise}`
    \end{cases}
    $$

    Choices for denominator include:

      - $\max(|x|, |y|)$: problem with $x=y=0$

      - $\mathrm{avg}(x, y)$: problem with $x = -y$; gives geometric mean

      - $\mathrm{avg}(|x|, |y|)$: problem when $x=y=0$

    Further reading:
      http://en.wikipedia.org/wiki/Relative_change_and_difference

    """

    a = array(a).flatten()
    b = array(b).flatten()
    #x = np.atleast_2d(array([a, b]))
    #fs = abs(x).max(axis=0)
    #fs[fs == 0] = 1
    #return (abs(x[0,:] - x[1,:]) / fs).flatten()

    es = []
    for x,y in zip(a,b):
        if x == y:  # covers inf, but not nan
            e = 0
        else:
            s = max(abs(x), abs(y))
            if s == 0:
                s = 1
            e = abs(x-y) / s
        es.append(e)
    return np.array(es)



def linf(a, b):
    return abs(a - b).max()


# Notes: I'm not using `scipy.spatial.distance.cosine` because it doesn't handle
# the zero norm cases.
def cosine(a, b):
    if not isfinite(a).all() or not isfinite(b).all():
        return np.nan
    A = norm(a)
    B = norm(b)
    if A == 0 and B == 0:
        return 1.0
    elif A != 0 and B != 0:
        return a.dot(b) / (A*B)
    else:
        return np.nan


def cumavg(x):
    """
    Cumulative average.

    >>> cumavg([1,2,3,4,5])
    array([1. , 1.5, 2. , 2.5, 3. ])

    """
    return cumsum(x) / arange(1.0, len(x)+1)


def normalize_zscore(data):
    """
    Shift and rescale data to be zero-mean and unit-variance along axis 0.
    """
    shift = data.mean(axis=0)
    rescale = data.std(axis=0)
    rescale[rescale == 0] = 1.0   # avoid divide by zero
    return (data - shift) / rescale


def normalize_interval(data):
    """
    Shift and rescale data so that it lies in the range [0,1].
    """
    shift = data.min(axis=0)
    rescale = data.ptp(axis=0)
    #rescale[rescale == 0] = 1.0  # avoid divide by zero
    x = (data - shift)
    x /= rescale
    return x


def mean_confidence_interval(a, confidence=0.95):
    "returns (mean, lower, upper)"
    a = asarray(a)
    n = a.shape[0]
    m = mean(a)
    h = a.std(ddof=1) / sqrt(n) * stats.t._ppf((1+confidence)/2., n-1)
    return m, m-h, m+h

def bernstein(samples, delta, R, check=True):
    """Plug-n-chug empirical Bernstein bound, computes "error bars" which hold with
    probability `1-δ` for the mean of independent samples from a given range
    `R=b-a` (known a priori).

    Returns ε such that the following bound hold,

      `p( mean(samples) - true_mean ≤ ε ) ≥ 1-δ`

    We assume that sample are independent (not necessarily identically
    distributed).

    Bound is based on sample variance `V` and a priori knowledge that RVs in are
    in the range `[a,b]` (although we only really require a known range `b-a`)

    The bound holds with probability `≥(1-δ)`.

    The sample mean has symmetric deviations so we get a two-sided bound by
    passing in 2*δ, i.e.,

      `p( |mean(samples) - true_mean| ≤ ε ) ≥ 1-2*δ`

    This is analogous to p-values, which make assumption of normally distributed
    random variables. This means that the bounds can be 'tighter', but the
    assumptions are usually not valid.

    """
    n = len(samples)
    if n <= 1: return np.nan
    V = np.var(samples, ddof=1)   # sample variance
    if check and np.ptp(samples) > R:
        raise ValueError(f'Sample range exceeded declared bound `{np.ptp(samples)} > {R}`')
    return sqrt(V*2*log(2.0/delta)/n) + R*(7.0/3.0)*log(2.0/delta)/(n-1)


def normalize(x, p=1):
    return x / norm(x, p)


def lidstone(p, delta):
    """
    Lidstone smoothing is a generalization of Laplace smoothing.
    """
    return normalize(p + delta)


@np.vectorize
def log1pexp(x):
    """
    Numerically stable implementation of log(1+exp(x)) aka softmax(0,x).

    -log1pexp(-x) is log(sigmoid(x))
    """
    if x <= -37:
        return exp(x)
    elif -37 <= x <= 18:
        return log1p(exp(x))
    elif 18 < x <= 33.3:
        return x + exp(-x)
    else:
        return x


@np.vectorize
def logsubexp(x, y):
    """
    Numerically stable computation of subtraction in log-space
    z = log(exp(x) - exp(y))
    """
    if x == y:
        return -np.inf
    elif x < y:
        return np.nan
    else:
        return x + log1mexp(y-x)


@np.vectorize
def log1mexp(x):
    """
    Numerically stable implementation of log(1-exp(x))

    Note: function is finite for x < 0.

    Source:
    http://cran.r-project.org/web/packages/Rmpfr/vignettes/log1mexp-note.pdf
    """
    if x >= 0:
        return np.nan
    else:
        a = abs(x)
        if 0 < a <= 0.693:
            return np.log(-np.expm1(-a))
        else:
            return np.log1p(-np.exp(-a))


def logmeanexp(xs):
    """
    Numerically stable implementation of log(mean(exp(xs))).

    Nptes:
      log(mean(exp(xs)))
      = log(sum(exp(xs))/n)
      = log(sum(exp(xs))) - log(n)
      = logsumexp(xs) - log(n)

    """
    return logsumexp(xs) - np.log(len(xs))


# based on implementation from scikits-learn
def logsumexp(arr, axis=None):
    """Computes the sum of arr assuming arr is in the log domain.

    Returns log(sum(exp(arr))) while minimizing the possibility of
    over/underflow.

    Examples:

    >>> a = arange(10)
    >>> log(sum(exp(a)))
    9.45862974442671

    >>> logsumexp(a)
    9.45862974442671

    >>> x = [[0, 0, 1000.0], [1000.0, 0, 0]]
    >>> logsumexp(x, axis=1)
    array([1000., 1000.])

    >>> logsumexp(x)
    1000.6931471805599

    >>> logsumexp(x, axis=0)
    array([1.00000000e+03, 6.93147181e-01, 1.00000000e+03])

    """
    arr = np.array(arr, dtype=np.double)
    if axis is None:
        arr = arr.ravel()
    else:
        arr = np.rollaxis(arr, axis)
    # Use the max to normalize, as with the log this is what accumulates the
    # less errors
    vmax = arr.max(axis=0)
    arr -= vmax
    exp(arr, out=arr)
    out = log(arr.sum(axis=0))
    out += vmax
    return out


def hardmax(x):
    p = np.zeros_like(x)
    p[x.argmax()] = 1
    return p


def simpless(w, B):
    """
    Reparameterization transform similar to softmax, but for the constraints set
      ${p | p ≥ 0, sum(p) ≤ B}.$
    """

    # Assuming B = 1,
    # pi = exp(xi) / 1 + sum_j exp(xj)
    #    = exp(xi - b) * exp(b) / exp(b)*exp(-b) + sum exp(xj - b) * exp(b)
    #    = exp(xi - b) / exp(-b) + sum_j exp(xj - b)

    b = w.max()
    e = np.exp(w - b)
    return B * e / (np.exp(-b) + e.sum())


def softmax(x, axis=None):
    """
    Compute softmax avoiding numerical overflow using [this
    trick](https://timvieira.github.io/blog/post/2014/02/11/exp-normalize-trick/).

    >>> x = [1, -10, 100, .5]
    >>> softmax(x)
    array([1.01122149e-43, 1.68891188e-48, 1.00000000e+00, 6.13336839e-44])

    >>> exp(x) / exp(x).sum()
    array([1.01122149e-43, 1.68891188e-48, 1.00000000e+00, 6.13336839e-44])

    >>> x = [[0, 0, 1000], [1000, 0, 0]]

    Normalize by row:

      >>> softmax(x, axis=0)
      array([[0. , 0.5, 1. ],
             [1. , 0.5, 0. ]])

    Normalize by column:

      >>> softmax(x, axis=1)
      array([[0., 0., 1.],
             [1., 0., 0.]])

    Normalize by cell:

      >>> softmax(x, axis=None)
      array([[0. , 0. , 0.5],
             [0.5, 0. , 0. ]])

    """
    a = np.array(x, dtype=np.double)
    if axis is None:
        v = a.ravel()
    else:
        v = np.rollaxis(a, axis)
    v -= v.max(axis=0)
    exp(v, out=v)
    v /= v.sum(axis=0)
    return a


# TODO: add support for the axis argument.
def d_softmax(out, x, adj):
    """
    out = softmax(x), adj are the adjoints we are chaining together.

      `    A(x) = logsumexp(x)`
      `∇  A(x) = softmax(x)`
      `∇² A(x) = d_softmax(x)`   % this method.

    """
    g = adj - adj @ out
    g *= out
    return g


# TODO: move elsewhere (test cases for this method live with my proximal
# operators).  This method should probably move there.
def project_onto_simplex(a, radius=1.0):
    """
    Project point a to the probability simplex.
    Returns (the projected point x, the zero threshold, and the residual value).

    """
    a = np.array(a)
    x0 = a.copy()
    d = len(x0)
    ind_sort = np.argsort(-x0)
    y0 = x0[ind_sort]
    ycum = np.cumsum(y0)
    val = 1.0/np.arange(1,d+1) * (ycum - radius)
    ind = np.nonzero(y0 > val)[0]
    rho = ind[-1]
    tau = val[rho]
    y = y0 - tau
    ind = np.nonzero(y < 0)
    y[ind] = 0
    x = x0.copy()
    x[ind_sort] = y
    return x, tau, .5*np.dot(x-a, x-a)


def project_pmin_simplex(p, pmin):
    q = pmin * np.ones_like(p)    # assumes all actions fire.
    c = 1
    while True:
        A0 = (pmin >= c * p)
        l = p[~A0].sum()
        if l == 0: break
        Δ = q[A0].sum()
        cc = (1 - Δ) / l
        if c == cc or not np.isfinite(cc): break
        c = cc
    return np.maximum(q, c * p)


log_of_2 = log(2)


# TODO: Create discrete distribution with an entropy method an make this a
# shortcut to the method on that instance.  Similarly for `kl_divergence` and
# `mutual information`, and `cross_entropy`.
def entropy(p):
    "Entropy of a discrete random variable with distribution `p`"
    assert len(p.shape) == 1
    p = p[p.nonzero()]
    return -p @ log(p) / log_of_2


def d_entropy(p):
    return -(1/np.log(2) + np.log2(p))


def kl_divergence(p, q):
    """ Compute KL divergence of two vectors, K(p || q).
    NOTE: If any value in q is 0.0 then the KL-divergence is infinite.
    """
    nz = p.nonzero()
    p = p[nz]
    q = q[nz]
    return p.dot(log(p) - log(q)) / log_of_2
#    return dot(p, log(p) - log(q)) / log_of_2


# KL(p||q) = sum_i p[i] log(p[i] / q[i])
#          = sum_i p[i] (log p[i] - log q[i])
#          = sum_i p[i] log p[i] - sum_i p[i] log(q[i])
#          = Entropy(p) + CrossEntropy(p,q)

def mutual_information(joint):
    """
    Mutual Information

    MI(x,y) = KL( p(x,y) || p(x) p(y) )

    We can compute this easily from the joint distribution

      joint = p(X=x,Y=y)

    because
      p(X=x) = sum_y p(X=x, Y=y)
      p(Y=y) = sum_x p(X=x, Y=y)

    relationships:
      MI(x,y) is the expected PMI(x,y) wrt p(x,y)
      MI(x,y) = KL(p(x,y) || p(x) p(y))

    properties:
      MI(X,Y) = MI(Y,X) is symmetric
    """
    # we can compute px and py from joint by applying the law of total probability
    px = joint.sum(axis=1)
    py = joint.sum(axis=0)
    independent = multiply.outer(px, py)
    assert joint.shape == independent.shape
    return kl_divergence(array(joint.flat), array(independent.flat))


def cross_entropy(p, q):
    """ Cross Entropy of two vectors,

    CE(p,q) = - \sum_i p[i] log q[i]

    Relationship to KL-divergence:

      CE(p,q) = entropy(p) + KL(p||q)
    """
    assert len(p) == len(q)
    p = p[p > 0]
    return -p @ log(q) / log_of_2


def assert_isdistr(p):
    assert (p >= 0).all()
    assert (p <= 1).all()
    assert abs(p.sum() - 1.0) < 0.000001


#def equal(a, b, tol=1e-10):
#    "L_{\inf}(a - b) < tol"
#    return inf_norm(a,b) < tol


def inf_norm(a, b):
    return abs(a - b).max()


# TODO: generalize to arrays? (or just use `compare`?)
def assert_equal(a, b, name='', verbose=False, throw=True, tol=0.001, color=1):
    """isfinite: asserts that *both* `a` and `b` must be finite.

    >>> assert_equal(0, 1, throw=0, color=0)
    0 1 err=[1.] fail

    >>> assert_equal(0, 0, verbose=1, color=0)
    0 0 err=[0] ok

    """
    try:
        assert not np.isnan(a).any() and not np.isnan(b).any()
    except (TypeError, AssertionError):
        pre = {"%s: " % name if name else ""}
        msg = f'{pre}want {a}, have {b}'
        if throw:
            raise AssertionError(msg)
        else:
            print(msg)

    a = np.asarray(a); b = np.asarray(b)
    err = relative_difference(a, b)
    if name:
        name = '%s: ' % name
    if verbose or np.any(err > tol):
        msg = f'{name}{a} {b} err={err}'
        if throw and err > tol:
            raise AssertionError(msg + ' ≥ tolerance (%s)' % tol)
        else:
            if not color:
                print(msg, 'ok' if err < tol else 'fail', sep=' ')
            else:
                print(msg, colors.green % 'ok' if err < tol else colors.red % 'fail', sep=' ')


def quadratic_formula(a,b,c):
    assert a != 0, 'this is a linear function!'
    return [
        (-b + np.sqrt(b**2 - 4*a*c)) / (2*a),
        (-b - np.sqrt(b**2 - 4*a*c)) / (2*a),
    ]


if __name__ == '__main__':

    def run_tests():

        from arsenal.maths import random_dist, fdcheck

        p = random_dist(30)
        fdcheck(lambda: entropy(p), p, d_entropy(p), throw=False)

        # To test the random state we'll run the function below from the same
        # random state using the restore_randome_state util.
        def foo():
            return [(random.randint(0, 100),
                     np.random.randint(0, 100))
                    for _ in range(10)]

        # Note that we have to run `a` and `b` in this order.
        with restore_random_state():
            a = foo()
        b = foo()
        assert a == b

        # Entropy tests
        assert entropy(array((0.5, 0.5))) == 1.0
        assert abs(entropy(array((0.75, 0.25))) - 0.8112781244) < 1e-10
        assert abs(entropy(array((0.1, 0.1, 0.8))) - 0.9219280948) < 1e-10

        # KL-divergence tests
        assert kl_divergence(array((0.5, 0.5)), array((0.5, 0.5))) == 0.0

        # KL, Entropy, and Cross Entropy relationship
        p = array([0.5, 0.5])
        q = array([0.4, 0.6])
        assert_equal(cross_entropy(p, q), (entropy(p) + kl_divergence(p, q)))

        # Normalize tests
        assert_equal(array([0.5, 0.5]), normalize(array([2, 2])))

        # Mutual Information tests
        #Pjoint = array([[0.25, 0.25], [0.25, 0.25]])

        Pjoint = normalize(np.random.rand(4,10))
        assert_equal(mutual_information(Pjoint),
                     mutual_information(Pjoint.T))

        def mi_independent_is_zero(px, py):
            assert mutual_information(multiply.outer(px, py)) <= 1e-10

        mi_independent_is_zero(p, q)
        mi_independent_is_zero(array([0.1, 0.1, 0.2, 0.6]),   # non-square joint
                               array([0.9, 0.1]))

        def test_softmax_grad():
            from arsenal.maths.checkgrad import fdcheck
            n = 20
            adj = np.random.uniform(-1,1,size=n)
            x = np.random.uniform(-1,1,size=n)
            out = softmax(x)
            g = d_softmax(out, x, adj)
            fdcheck(lambda: softmax(x).dot(adj), x, g)

        test_softmax_grad()

        assert relative_difference(np.inf, np.inf) == 0.0

        print('passed tests.')

    run_tests()

    import doctest
    doctest.testmod()



from arsenal.nlp.annotation import bio2span, Span

def test_bio2span():

    tests = [
        (['I-NUM','I-TEMP'], [Span(label='NUM', begins=0, ends=1), Span(label='TEMP', begins=1, ends=2)]),
        (['I-NUM','B-TEMP'], [Span(label='NUM', begins=0, ends=1), Span(label='TEMP', begins=1, ends=2)]),
        (['B-NUM','B-TEMP'], [Span(label='NUM', begins=0, ends=1), Span(label='TEMP', begins=1, ends=2)]),
        (['B-NUM','B-TEMP'], [Span(label='NUM', begins=0, ends=1), Span(label='TEMP', begins=1, ends=2)]),
        (['B-NUM','O'], [Span(label='NUM', begins=0, ends=1)]),
        (['O','B-NUM'], [Span(label='NUM', begins=1, ends=2)]),
        (['O','B-NUM','O'], [Span(label='NUM', begins=1, ends=2)]),
        (['O','B-NUM','I-NUM'], [Span(label='NUM', begins=1, ends=3)]),
        (
            ['O', 'O', 'O',
             'I-NUM', 'I-NUM', 'I-NUM',
             'I-TEMP', 'I-TEMP', 'I-TEMP', 'I-TEMP', 'I-TEMP', 'I-TEMP', 'I-TEMP', 'I-TEMP',
             'O',
             'I-NUM', 'I-NUM', 'I-NUM', 'I-NUM', 'I-NUM', 'I-NUM', 'I-NUM', 'I-NUM',
             'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O'],
            [Span(label='NUM', begins=3, ends=6),
             Span(label='TEMP', begins=6, ends=14),
             Span(label='NUM', begins=15, ends=23)
             ]
        ),
        (
            ['O', 'O', 'O', 'O', 'O', 'O',
             'I-TEMP',
             'I-NUM', 'I-NUM', 'I-NUM',
             'O',
             'I-NUM', 'I-NUM', 'I-NUM', 'I-NUM',
             'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O',
             'I-TEMP', 'I-TEMP', 'I-TEMP',
             'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O'],
            [Span(label='TEMP', begins=6, ends=7),
             Span(label='NUM', begins=7, ends=10),
             Span(label='NUM', begins=11, ends=15),
             Span(label='TEMP', begins=23, ends=26),
             ]
        ),
        (
            ['O', 'O', 'O',
             'B-NUM', 'I-NUM', 'I-NUM', 'I-NUM' ,
             'B-NUM', 'I-NUM', 'I-NUM', 'I-NUM',
             'O',
             'B-NUM', 'I-NUM', 'I-NUM',
             'O', 'O', 'O',
             'B-TEMP', 'I-TEMP',
             'O'],
            [Span(label='NUM', begins=3, ends=7),
             Span(label='NUM', begins=7, ends=11),
             Span(label='NUM', begins=12, ends=15),
             Span(label='TEMP', begins=18, ends=20),
             ]
        ),
    ]

    for x, want in tests:
        print(x)
        have = bio2span(x, include_O=False)
        if want == have:
            print('\033[32mpassed\033[0m')
        else:
            print('\033[31mfailed:\033[0m')
            print('  want:', want)
            print('       have:', have)
        print()


    print('*************************************************************************')
    print('** Including O Spans.')

    include_O = [
        (['O','B-NUM','I-DATE'],
         [Span(label='O', begins=0, ends=1),
          Span(label='NUM', begins=1, ends=2),
          Span(label='DATE', begins=2, ends=3)]
        ),
        (['O','B-NUM','O','I-DATE'],
         [Span(label='O', begins=0, ends=1), Span(label='NUM', begins=1, ends=2), Span(label='O', begins=2, ends=3),
          Span(label='DATE', begins=3, ends=4)]
        ),
        (['O','B-NUM','O','I-DATE', 'O', 'O'],
         [Span(label='O', begins=0, ends=1),
          Span(label='NUM', begins=1, ends=2),
          Span(label='O', begins=2, ends=3),
          Span(label='DATE', begins=3, ends=4),
          Span(label='O', begins=4, ends=5),
          Span(label='O', begins=5, ends=6)]
         ),
        (['O'],
         [Span(label='O', begins=0, ends=1)]
        ),
    ]

    for x, want in include_O:
        print(x)
        have = bio2span(x, include_O=True)
        if want == have:
            print('\033[32mpassed\033[0m')
        else:
            print('\033[31mfailed:\033[0m')
            print('  want:', want)
            print('  have:', have)
            raise AssertionError
        print()

    print('passed test_bio2span')


if __name__ == '__main__':
    test_bio2span()



import numpy as np
from random import random, choice
from arsenal.datastructures.heap import MaxHeap, LocatorMaxHeap, MinMaxHeap, BoundedMaxHeap


class SlowPriorityQueue:
    "Priority queue; Prioritization is based on `__lt__`."
    def __init__(self, maxsize=0):
        self.maxsize = maxsize
        self.items = {}
    def __iter__(self):
        return iter(self.items)
    def __setitem__(self, k, v):
        self.items[k] = v
        if self.maxsize > 0 and len(self.items) > self.maxsize:
            self.popmin()
        if self.maxsize > 0:
            assert len(self.items) <= self.maxsize
    def popmin(self):
        v,k = min((v, k) for k, v in self.items.items())
        del self.items[k]
        return k, v
    def pop(self):
        v,k = max((v, k) for k, v in self.items.items())
        del self.items[k]
        return k, v
    def __len__(self):
        return len(self.items)


class SlowLocatorHeap:
    "Priority queue; Prioritization is based on `__lt__`."
    def __init__(self):
        self.items = {}
    def __setitem__(self, k, v):
        self.items[k] = v
    def popmin(self):
        v,k = min((v,k) for k,v in self.items.items())
        del self.items[k]
        return k, v
    def popmax(self):
        v,k = max((v,k) for k,v in self.items.items())
        del self.items[k]
        return k, v
    def __len__(self):
        return len(self.items)


def test_basic():
    L = MaxHeap(cap=1)
    S = SlowPriorityQueue()

    for t in range(1000):

        if np.random.uniform(0, 1) < .5:
            v = np.random.uniform(-1, 1)
            L.push(v)
            S[t] = v
            L.check()

        else:
            if L or S:
                v1 = L.pop()
                _, v2 = S.pop()
                assert v1 == v2, [v1, v2]

        L.check()

    print('[basic random workload] pass.')


def test_locator():
    L = LocatorMaxHeap(cap=1)
    S = SlowLocatorHeap()

    K = 1
    for _ in range(1000):

        if np.random.uniform(0, 1) < .5:

            if np.random.uniform(0, 1) < .1:  # new key:
                K += 1
                k = K
            else:
                k = int(np.random.randint(K))

            k = f'mykey[{k}]'

            v = np.random.uniform(-1, 1)
            L[k] = v
            S[k] = v
            L.check()

            #print('upsert', k, v)

        else:

            A = S.items
            B = {k: L[k] for k in L.loc}

            #print('want:', A)
            #if A != B: print('got: ', B)
            assert A == B

            if L or S:
                v2 = S.popmax()
                #print('pop', v2)
                v1 = L.pop()

                assert v1 == v2, [v1, v2]

        L.check()

    print('[locator random workload] pass.')


def test_minmax():
    L = MinMaxHeap(cap=1)
    S = SlowLocatorHeap()

    K = 1
    for _ in range(1000):

        if np.random.uniform(0, 1) < .5:

            if np.random.uniform(0, 1) < .1:  # new key:
                K += 1
                k = K
            else:
                k = int(np.random.randint(K))

            k = f'mykey[{k}]'

            v = np.random.uniform(-1, 1)
            L[k] = v
            S[k] = v

            #print('upsert', k, v)

        else:

            A = S.items
            B = L.map()

            #print('want:', A)
            #if A != B: print('got: ', B)
            assert A == B

            if L or S:

                if np.random.uniform(0, 1) < .5:
                    v2 = S.popmin()
                    v1 = L.popmin()
                else:
                    v2 = S.popmax()
                    v1 = L.popmax()

                assert v1 == v2, [v1, v2]

        L.check()

    print('[min-max random workload] pass.')



def test_bounded():
    maxsize = 5
    L = BoundedMaxHeap(maxsize=maxsize, cap=1)
    S = SlowPriorityQueue(maxsize=maxsize)

    K = 1
    for _ in range(1000):

        if np.random.uniform(0, 1) < .5:

            if np.random.uniform(0, 1) < .1:  # new key:
                K += 1
                k = K
            else:
                k = int(np.random.randint(K))

            k = f'mykey[{k}]'

            v = np.random.uniform(-1, 1)
            L[k] = v
            S[k] = v
            L.check()

            #print('upsert', k, v)

        else:

            A = S.items
            B = L.map()

            #print('want:', A)
            #if A != B: print('got: ', B)
            assert A == B

            if L or S:
                v2 = S.pop()
                v1 = L.pop()
                assert v1 == v2, [v1, v2]

        L.check()

    print('[bounded random workload] pass.')


if __name__ == '__main__':
    from arsenal import testing_framework
    testing_framework(globals())



from heapq import heapify, heappop, heappush

# This implementation is based on
# https://code.activestate.com/recipes/522995-priority-dict-a-priority-queue-with-updatable-prio/

class pdict(dict):
    """Dictionary that can be used as a priority queue.

    Keys of the dictionary are items to be put into the queue, and values
    are their respective priorities. All dictionary methods work as expected.
    The advantage over a standard heapq-based priority queue is
    that priorities of items can be efficiently updated (amortized O(1))
    using code as 'thedict[item] = new_priority.'

    The 'smallest' method can be used to return the object with lowest
    priority, and 'pop_smallest' also removes it.

    The 'sorted_iter' method provides a destructive sorted iterator.
    """

    def __init__(self, *args, **kwargs):
        super(pdict, self).__init__(*args, **kwargs)
        self._rebuild_heap()

    def _rebuild_heap(self):
        self._heap = [(v, k) for k, v in self.items()]
        heapify(self._heap)

    def pop(self):
        return self.popitem()[0]

    def popitem(self):
        """Return the item with the lowest priority.

        Raises IndexError if the object is empty.
        """

        heap = self._heap
        v, k = heap[0]
        while k not in self or self[k] != v:
            heappop(heap)
            v, k = heap[0]
        del self[k]
        return (k, v)

    def __setitem__(self, key, val):
        # We are not going to remove the previous value from the heap, since
        # this would have a cost O(n).
        super(pdict, self).__setitem__(key, val)
        heappush(self._heap, (val, key))
        return self

    def setdefault(self, key, val):
        if key not in self:
            self[key] = val
            return val
        return self[key]

    def update(self, *args, **kwargs):
        # Reimplementing dict.update is tricky -- see e.g.
        # http://mail.python.org/pipermail/python-ideas/2007-May/000744.html
        # We just rebuild the heap from scratch after passing to super.

        super(pdict, self).update(*args, **kwargs)
        self._rebuild_heap()

    def sorted_iter(self):
        """Sorted iterator of the priority dictionary items.

        Beware: this will destroy elements as they are returned.
        """

        while self:
            yield self.pop_smallest()



import numpy as np
from matplotlib import pyplot as pl
import pandas
from collections import defaultdict
from time import time

from arsenal.viz.util import update_ax
from arsenal.misc import ddict

class LearningCurve:
    """
    Plot learning curve as data arrives.
    """

    def __init__(self, name, sty=None, legend=True, ax=None):
        self.name = name
        self.baselines = {}
        self.data = defaultdict(list)
        self.sty = defaultdict(dict)
        if sty is not None:
            self.sty.update(sty)

        self.ax = pl.figure(figsize=(10,6)).add_subplot(111) if ax is None else ax
        self.legend = legend
        self.yscale = None
        self.xscale = None

        self.last_update = time()
        self.min_time = 0.5

        self.smoothing = None
        self._bands = None
        self.widget = None

    def smooth(self, type, aggregate, **kwargs):
        # TODO: look at notes:interpolated-signal for additional ideas for
        # smoothing the learning curve.
        if type == 'rolling':
            assert 'window' in kwargs
        elif type == 'ewm':
            assert 'half_life' in kwargs
        else:
            raise ValueError(self.smoothing.get('type'))
        self.smoothing = dict(type=type, aggregate=aggregate, **kwargs)
        self.add_widget()
        return self

    def bands(self, type):
        assert self.smoothing is not None
        self._bands = dict(type=type)
        return self

    def loglog(self):
        self.xscale = 'log'
        self.yscale = 'log'
        return self

    def semilogy(self):
        self.yscale = 'log'
        return self

    def semilogx(self):
        self.xscale = 'log'
        return self

    def draw(self):
        ax = self.ax
        with update_ax(ax):
            sty = self.sty
            for k, v in self.baselines.items():
                ax.axhline(v, label=k, **sty[k])
            data = self.data
            for k, v in data.items():
                xs, ys = np.array(data[k]).T

                if self.smoothing is not None:
                    # mute the raw signal when we are smoothing.
                    sty[k]['alpha'] = 0.5

                [l] = ax.plot(xs, ys, label=k, **sty[k])
                c = l.get_color()
                # show a dot in addition to the line
                if 0:
                    s = sty[k].copy()
                    s['lw'] = 0
                    s['c'] = c
                    ax.scatter(xs, ys, label=k, **s)

                self.draw_smoothing(xs, ys, c=c)
                self.draw_bands(xs, ys, c=c)

            if self.xscale: ax.set_xscale(self.xscale)
            if self.yscale: ax.set_yscale(self.yscale)
            if self.name:   ax.set_title(self.name)
            if self.legend: ax.legend(loc='best')

            ax = self.ax
            ax.spines['top'].set_visible(False)
            ax.spines['right'].set_visible(False)
            #ax.spines['bottom'].set_visible(False)
            #ax.spines['left'].set_visible(False)

            self.draw_extra(ax)
            #self.ax.figure.tight_layout()

        return self

    def add_widget(self):
        if self.widget is not None: return
        from matplotlib.widgets import TextBox
        ax_widget = self.ax.figure.add_axes([0.1, .93, 0.06, 0.037])  # [left, bottom, width, height]

        def submit(text):
            if not text: return
            try:
                x = float(text)
                self.smoothing['half_life'] = x
                assert x > 0
            except (ValueError, AssertionError):
                print('bad value for smooting parameter')

        self.widget = TextBox(ax_widget, 'Smoothing ',
                              initial = str(self.smoothing['half_life']))

        self.widget.on_submit(submit)

    def draw_extra(self, ax):
        return

    def smoothed_signal(self, xs, ys):
        assert self.smoothing is not None
        s = pandas.Series(ys)
        if self.smoothing['type'] == 'ewm':
            # TODO: this doesn't use the distances in xs
            return s.ewm(halflife=self.smoothing['half_life'])
        elif self.smoothing['type'] == 'rolling':
            window = min(len(ys), self.smoothing['window'])
            return s.rolling(window, min_periods=0)

    def draw_smoothing(self, xs, ys, c):
        # Note: the smoothing happens on the original signal, not in log space.
        # We should probably add an option to smooth in log/log-log space.
        if self.smoothing is None: return
        r = self.smoothed_signal(xs, ys)
        if self.smoothing['aggregate'] == 'mean':
            zs = r.mean()
        if self.smoothing['aggregate'] == 'median':
            zs = r.median()
        self.ax.plot(xs, zs, lw=2, c=c)

    def draw_bands(self, xs, ys, c):
        if self._bands is None: return

        r = self.smoothed_signal(xs, ys)

        if self._bands['type'] == 'std':
            M = r.mean()
            s = r.std()
            U = M + 2*s
            L = M - 2*s

        if self._bands['type'] == 'quantile':
            U = r.quantile(.9)
            L = r.quantile(.1)

        self.ax.fill_between(xs, U, L, color=c, alpha=0.25)

    def update(self, iteration, **kwargs):
        "Update plots, if ``iteration is None`` we'll use ``iteration=len(data)``"
        return self._update(iteration, kwargs)

    def _update(self, iteration, kwargs):
        "Update plots, if ``iteration is None`` we'll use ``iteration=len(data)``"
        assert isinstance(kwargs, dict)
        data = self.data
        for k, v in kwargs.items():
            i = len(data[k]) if iteration is None else iteration
            data[k].append([i, v])
        if self.should_update():
            self.draw()
            self.last_update = time()
        return self

    def should_update(self):
        "Returns true if its been long enough (>= `min_time`) since the `last_update`."
        return time() - self.last_update >= self.min_time

    def __reduce__(self):
        # Default pickle fails because of the reference to the plotting axis
        x = self.__dict__.copy()
        x['ax'] = None
        return (LearningCurve, (self.name, self.sty), x)


lc = ddict(LearningCurve)



#def run():
#
#    lc = LearningCurve('test')
#    lc.smooth('ewm', 'mean', half_life = 0.01).bands('std')
#    for t in range(1, 1000):
#        lc.update(t, signal = np.exp(np.log(t) * -0.5 + np.random.randn()))
#        if t % 10 == 0:
#            lc.update(t, signal2 = np.exp(np.log(t) * -0.25 + np.random.randn()))
#        lc.loglog().draw()
#
#    pl.ioff(); pl.show()
#
#
#if __name__ == '__main__':
#    run()
#



#from importlib import import_module
from arsenal.iterview import iterview
from arsenal.iterextras import *
from arsenal.terminal import *
from arsenal import terminal as colors
from arsenal.alphabet import Alphabet
from arsenal.integerizer import Integerizer
from arsenal.timer import Timer, timers, timeit
#from arsenal.viz import axman, update_ax
#from arsenal import math
#from arsenal.math import wide_dataframe
#from arsenal.debug import ip
#from arsenal.humanreadable import htime
from arsenal.misc import ddict
from arsenal.assertions import assert_throws

import platform
if platform.system() != 'Windows':
    from arsenal.robust import timelimit, Timeout

from arsenal.maths import restore_random_state
from arsenal.datastructures import *
from arsenal.testing_framework import *
from arsenal.datastructures.multimap import MultiMap
from arsenal.profiling import profiler



"""
Utilities for working with the Cholesky factorization of a positive-definite matrix.
"""
import numpy as np
import scipy.linalg as la
#import numpy.linalg as la
from arsenal.maths import blocks, random_psd


class Cholesky:
    """Interface to Cholesky factorization of a matrix, which supports updates
    (growing and rank-one updates) as well as efficient utility methods (e.g.,
    solve, det).

    Remarks:

     - We store the factorization (`self.L`) as upper triangular to be
       consistent with `scipy.linalg.cholesky`.  Note that
       `numpy.linalg.cholesky` is lower triangular for some reason.

    """

    def __init__(self, M):
        self.L = la.cholesky(M)

    def update_rank_one(self, x, x_copy=True):
        """
        Rank-one update: compute Chol(M + x xᵀ) given L = Chol(M).

        Remarks:
         - Update is performed in-place (so `self.L` is mutated).
         - We don't need to know `M` to peform the update.
        """
        if x_copy: x = x.copy()
        L = self.L.T
        [n] = x.shape
        assert L.shape == (n,n)
        for k in range(n):
            r = np.hypot(L[k, k], x[k])      #r = np.sqrt(L[k, k]**2 + x[k]**2)
            c = r / L[k, k]
            s = x[k] / L[k, k]
            L[k, k]    = r
            L[k+1:, k] = (L[k+1:n, k] + s * x[k+1:n]) / c
            x[k+1:]    = c * x[k+1:n] - s * L[k+1:n, k]

    def downdate_rank_one(self, z):
        """
        Rank-one down-date: compute Chol(M - x xᵀ) given L = Chol(M).
        """
        # based on https://github.com/scipy/scipy/issues/8188
        L = self.L
        n = L.shape[0]
        eps = n * np.spacing(1.)  # For complex this needs modification
        alpha, beta = np.empty_like(z), np.empty_like(z)
        alpha[-1], beta[-1] = 1., 1.

        for r in range(n):
            a = z[r] / L[r, r]
            alpha[r] = alpha[r-1] - a ** 2
            # Numerically zero or negative
            if alpha[r] < eps:
                # Made up err msg.
                raise la.LinAlgError('The Cholesky factor becomes nonpositive'
                                     'with this downdate at the step {}'.format(r))
            beta[r] = np.sqrt(alpha[r])
            z[r+1:] -= a * L[r, r+1:]
            L[r, r:] *= beta[r] / beta[r-1]
            L[r, r+1:] -= a/(beta[r] * beta[r-1]) * z[r+1:]

    def update_grow(self, B, D):
        """
        Growth update: compute `Chol([[  A, B], [B.T, D]])` given `L = Chol(A)`

        Remarks:
         - We don't need to know `A` to peform the update.

        """
        # TODO: amortize the cost of growing the matrix by using a doubling strategy?
        L = self.L
        [n, m] = B.shape
        assert L.shape == (n,n) and B.shape == (n, m) and D.shape == (m, m)
        B1 = la.solve_triangular(L, B, trans=True)
        C1 = la.cholesky(D - B1.T @ B1)
        L1 = np.block([[L,                B1],
                       [np.zeros((m, n)), C1]])
        self.L = L1

    def solve(self, b):
        return la.cho_solve((self.L, False), b)

    def det(self):
        return np.prod(np.diag(self.L))**2


def test_grow():

    n = 9
    m = 4

    X = random_psd(n)

    [A,B], [C,D] = blocks(X, n-m)
    assert np.allclose(A, A.T)
    assert np.allclose(B, C.T)
    assert np.allclose(D, D.T)

    L = Cholesky(A)
    L.update_grow(B, D)

    assert np.allclose(la.cholesky(X), L.L)
    print('[test grow] pass!')


def test_rank_one():
    n = 10

    A = random_psd(n)
    x = np.random.randn(n)

    L = Cholesky(A)
    L.update_rank_one(x)
    assert np.allclose(L.L, la.cholesky(A + np.outer(x, x)))

    L.downdate_rank_one(x)
    assert np.allclose(L.L, la.cholesky(A))

    print('[test rank-one] pass!')


def test_util():
    n = 10
    X = random_psd(n)

    L = Cholesky(X)
    assert np.allclose(L.det(), np.linalg.det(X))

    # use Cholesky factorization to solve a linear system in O(n^2)
    b = np.random.randn(n)
    assert np.allclose(L.solve(b), la.solve(X, b))

    print('[test util] pass!')


if __name__ == '__main__':
    test_rank_one()
    test_grow()
    test_util()



"""Caching with pickle and the file system.

"""
import pickle
from path import Path as path
from arsenal.fsutils import filesize
from arsenal.timer import timeit


def load(filename, default=None, saveit=False, verbose=False):
    "Load cached item by `name`, on miss call `get` function and cached results."
    f = path(filename)
    if f.exists():
        if verbose:
            print('[load] %s, size = %s' % (f, filesize(f)))
            with timeit('[load] %s' % filename):
                with open(f) as pkl:
                    return pickle.load(pkl)
        else:
            with open(f) as pkl:
                return pickle.load(pkl)
    else:
        if default is None:
            raise OSError("File not found '%s'" % filename)
        with timeit('[load] make %s' % filename):
            val = default()
        if saveit:
            save(filename, val, verbose=verbose)
        return val


def save(filename, val, verbose=False):
    "Save `val` so we can load it via `load`."
    if verbose:
        with timeit('[save] %s' % filename):
            with open(filename, 'w') as pkl:
                pickle.dump(val, pkl)
            print('[save] %s, size = %s' % (filename, filesize(filename)))
    else:
        with open(filename, 'w') as pkl:
            pickle.dump(val, pkl)
    return val



from os import environ
DISPLAY = True
if not environ.get('DISPLAY'):
    import matplotlib
    #print 'Not a display environment.'
    matplotlib.use('Agg')
    DISPLAY = False

import pandas as pd, numpy as np
import matplotlib.pyplot as pl
#from sys import stderr
from collections import defaultdict
from contextlib import contextmanager
from matplotlib.backends.backend_pdf import PdfPages
from mpl_toolkits.mplot3d import Axes3D
from arsenal import colors
#from arsenal.viz.covariance_ellipse import covariance_ellipse
from arsenal.misc import ddict


#from palettable.colorbrewer import qualitative


simple_palette = ['r','g','b','y','c','m','k']
#default_palette = np.array(qualitative.Set1_6.mpl_colors)


#def name2color(palette = default_palette):
def name2color(palette = simple_palette):
    "Create a mapping from names to matplotlib colors."
    palette = list(palette)
    i = -1
    n = len(palette)
    def next_color():
        nonlocal i
        i += 1
        return palette[i % n]
    return defaultdict(next_color)


def save_plots(pdf):
    "save all plots to pdf"
    pp = PdfPages(pdf)
    for i in pl.get_fignums():
        pl.figure(i)
        pl.savefig(pp, format='pdf')
    pp.close()
    print(colors.yellow % 'saved plots to "%s"' % pdf)


# global reference to all of the plots
def newax():
    return pl.figure().add_subplot(111)
AX = defaultdict(newax)
DATA = defaultdict(list)


# TODO: [2018-04-18 Wed] Add support for visualizing constraints. Apparently,
# I'm constantly plotting constraints these days.
#
#  - There may be better ways to implement this (better defaults at least)
#
#    - Look into rasbt's decision boundary plotting script for some example usage of all these things.
#       https://github.com/rasbt/mlxtend/blob/62aea2a9fb6fafdecedfa041a2121c002e47dac9/mlxtend/plotting/decision_regions.py
#
#    - figure out the difference between contoutr/contourf)
#
#    - Look into value interpolation strategies
#      https://matplotlib.org/gallery/images_contours_and_fields/triinterp_demo.html#sphx-glr-gallery-images-contours-and-fields-triinterp-demo-py.
#
#    - For constraints look at value mask:
#      https://matplotlib.org/gallery/images_contours_and_fields/contour_corner_mask.html#sphx-glr-gallery-images-contours-and-fields-contour-corner-mask-py
#
# - For rendering constraints, we might want to use hatching instead of
#   something opaque.
#
def contour_plot(f, xdomain, ydomain, color='viridis', alpha=0.5, levels=None, ax=None):
    "Contour plot of a function of two variables."
    from arsenal import iterview
    if ax is None: ax = pl.gca()
    [xmin, xmax, _] = xdomain; [ymin, ymax, _] = ydomain
    X, Y = np.meshgrid(np.linspace(*xdomain), np.linspace(*ydomain))
    Z = np.array([f(np.array([x,y])) for (x,y) in iterview(zip(X.flat, Y.flat), length=len(X.flat))]).reshape(X.shape)
    contours = ax.contour(X, Y, Z, 20, colors='black', levels=levels)
    ax.clabel(contours, inline=True, fontsize=8)
    if color is not None:
        ax.imshow(Z, extent=[xmin, xmax, ymin, ymax], origin='lower', cmap=color, alpha=alpha)
        #ax.axis(aspect='scalar')
    ax.figure.tight_layout()
    ax.set_xlim(xmin,xmax); ax.set_ylim(ymin,ymax)

# TODO: No need to say "plot" we're already in a module called "viz". The whole
# point is reduce clutter when plotting.
contour = contour_plot

# TODO: Create an alias which case-analyzes and plots 3d vs 2d accordingly?
# TODO: also support interactive sliders and animation for when there are more parameters. use the same range notation.
def plot3d(f, xdomain, ydomain, ax=None):
    "3d surface plot of a function of two variables."
    #[xmin, xmax, _] = xdomain; [ymin, ymax, _] = ydomain
    X, Y = np.meshgrid(np.linspace(*xdomain), np.linspace(*ydomain))
    Z = np.array([f(np.array([x,y])) for (x,y) in zip(X.flat, Y.flat)]).reshape(X.shape)
    ax = pl.figure().gca(projection='3d') if ax is None else ax
    ax.plot_surface(X, Y, Z, cmap='viridis', linewidth=0, antialiased=True)
    return ax


class plot_xsection:
    def __init__(self, f, a, b, n, ax=None, opts=None):
        """
        Plot a cross section of `f` by interpolating from `x0 to `x1` by `n`
        evenly space points.
        """
        if opts is None: opts = {}
        if ax is None: ax = pl.gca()
        self.n = n
        self.a = a
        self.b = b
        self.ts = np.linspace(0,1,n)
        self.fs = [f(xt) for xt in self.curve()]
        ax.plot(self.ts, self.fs, **opts)
        ax.set_xlabel('interpolation coefficient')
        self.ax = ax

    def __call__(self, f, opts=None):
        return plot_xsection(f=f, a=self.a, b=self.b, n=self.n, ax=self.ax, opts=opts)

    def curve(self):
        "Sweep a curve in parameter spaces which is convex combination of `a` and `b`."
        for t in self.ts:
            yield self.a*(1-t) + self.b*t


class NumericalDebug:
    """Incrementally builds a DataFrame, includes plotting and comparison method.

    The quickest way to use it is

      >>> from arsenal.viz import DEBUG
      >>> d = DEBUG['test1']
      >>> d.update(want=1, have=1)
      >>> d.update(want=1, have=1.01)
      >>> d.update(want=1, have=0.99)
         want  have
      0     1  1.00
      1     1  1.01
      2     1  0.99

    To plots and runs numerical comparison tests,

      >>> d.compare()     # doctest: +SKIP

    """

    def __init__(self, name):
        self.name = name
        self._data = []
        self._df = None
        self.ax = None
        self.uptodate = True

    @property
    def df(self):
        "lazily make DataFrame from _data."
        if not self.uptodate:
            self._df = pd.DataFrame(self._data)
        self.uptodate = True
        return self._df

    def update(self, **kw):
        "Pass in column values for the row by name as keyword arguments."
        self._data.append(kw)
        self.uptodate = False
        return self

    def compare(self, want='want', have='have', show_regression=1, scatter=1, **kw):
        from arsenal.maths import compare
        if self.ax is None:
            self.ax = pl.figure().add_subplot(111)
        if self.df.empty:
            return
        with update_ax(self.ax):
            compare(want, have, data=self.df).plot(ax=self.ax, **kw)

# Global references to numerical debugger class.
DEBUG = ddict(NumericalDebug)

@contextmanager
def lineplot(name, with_ax=False, halflife=20, xlabel=None, ylabel=None, title=None, **style):
    with axman(name, xlabel=xlabel, ylabel=ylabel, title=title) as ax:
        data = DATA[name]
        if with_ax:
            yield (data, ax)
        else:
            yield data
        ax.plot(list(range(len(data))), data, alpha=0.5, **style)
        if halflife:
            ax.plot(pd.Series(data).ewm(halflife=halflife).mean(), alpha=0.5, c='k', lw=2)


@contextmanager
def axman(name, xlabel=None, ylabel=None, title=None, clear=True):
    """`axman` is axis manager. Manages clearing, updating and maintaining a global
    handle to a named plot.

    """
    ax = AX[name]
    prev_ax = pl.gca()
    with update_ax(ax, clear=clear):
        _try_sca(ax)
        yield ax
        if xlabel:
            ax.set_xlabel(xlabel)
        if ylabel:
            ax.set_ylabel(ylabel)
        ax.set_title(title or name)  # `title` overrides `name`.
        #ax.figure.tight_layout()
        _try_sca(prev_ax)


def _try_sca(ax):
    try:
        pl.sca(ax)
    except ValueError:
        pass


@contextmanager
def update_ax(ax, clear=True):
    "Manages clearing and updating a plot."
    if not hasattr(ax, '_did_show'):
        ax._did_show = False
    if clear:
        ax.clear()
    yield
    for _ in range(2):
        try:
            ax.figure.canvas.draw_idle()
            ax.figure.canvas.flush_events()
            if not ax._did_show:
                pl.show(block=False)
                ax._did_show = True
        except (NotImplementedError, AttributeError):
            #print >> stderr, 'warning failed to update plot.'
            pass


@contextmanager
def scatter_manager(name, with_ax=False, xlabel=None, ylabel=None, title=None, **style):
    with axman(name, xlabel=xlabel, ylabel=ylabel, title=title) as ax:
        data = DATA[name]
        if with_ax:
            yield (data, ax)
        else:
            yield data
        x,y = list(zip(*data))
        ax.scatter(x, y, alpha=0.5, lw=0, **style)


def test():
    d = DEBUG['test1']
    d.update(want=1, have=1)
    d.update(want=1, have=1.01)
    d.update(want=1, have=0.99)
    print(d.df)


if __name__ == '__main__':
    test()
